#!/usr/bin/env python
#
# Copyright (c), 2018-2021, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
import math
from xml.etree import ElementTree
from elementpath.helpers import days_from_common_era, months2days, \
    round_number, is_idrefs, collapse_white_spaces, escape_json_string, \
    get_double, numeric_equal, numeric_not_equal, equal, not_equal, \
    match_wildcard, unescape_json_string, iter_sequence, split_function_test


class HelperFunctionsTest(unittest.TestCase):

    def test_node_is_idref_function(self):
        self.assertTrue(is_idrefs(ElementTree.XML('<A>xyz</A>').text))
        self.assertTrue(is_idrefs(ElementTree.XML('<A>xyz abc</A>').text))
        self.assertFalse(is_idrefs(ElementTree.XML('<A>12345</A>').text))
        self.assertTrue(is_idrefs('alpha'))
        self.assertTrue(is_idrefs('alpha beta'))
        self.assertFalse(is_idrefs('12345'))

    def test_days_from_common_era_function(self):
        days4y = 365 * 3 + 366
        days100y = days4y * 24 + 365 * 4
        days400y = days100y * 4 + 1

        self.assertEqual(days_from_common_era(0), 0)
        self.assertEqual(days_from_common_era(1), 365)
        self.assertEqual(days_from_common_era(3), 365 * 3)
        self.assertEqual(days_from_common_era(4), days4y)
        self.assertEqual(days_from_common_era(100), days100y)
        self.assertEqual(days_from_common_era(200), days100y * 2)
        self.assertEqual(days_from_common_era(300), days100y * 3)
        self.assertEqual(days_from_common_era(400), days400y)
        self.assertEqual(days_from_common_era(800), 2 * days400y)
        self.assertEqual(days_from_common_era(-1), -366)
        self.assertEqual(days_from_common_era(-4), -days4y)
        self.assertEqual(days_from_common_era(-5), -days4y - 366)
        self.assertEqual(days_from_common_era(-100), -days100y - 1)
        self.assertEqual(days_from_common_era(-200), -days100y * 2 - 1)
        self.assertEqual(days_from_common_era(-300), -days100y * 3 - 1)
        self.assertEqual(days_from_common_era(-101), -days100y - 366)
        self.assertEqual(days_from_common_era(-400), -days400y)
        self.assertEqual(days_from_common_era(-401), -days400y - 366)
        self.assertEqual(days_from_common_era(-800), -days400y * 2)

    def test_months2days_function(self):
        self.assertEqual(months2days(-119, 1, 12 * 319), 116512)
        self.assertEqual(months2days(200, 1, -12 * 320) - 1, -116877 - 2)

        # 0000 BCE tests
        self.assertEqual(months2days(0, 1, 12), 366)
        self.assertEqual(months2days(0, 1, -12), -365)
        self.assertEqual(months2days(1, 1, 12), 365)
        self.assertEqual(months2days(1, 1, -12), -366)

        # xs:duration ordering related tests
        self.assertEqual(months2days(year=1696, month=9, months_delta=0), 0)
        self.assertEqual(months2days(1696, 9, 1), 30)
        self.assertEqual(months2days(1696, 9, 2), 61)
        self.assertEqual(months2days(1696, 9, 3), 91)
        self.assertEqual(months2days(1696, 9, 4), 122)
        self.assertEqual(months2days(1696, 9, 5), 153)
        self.assertEqual(months2days(1696, 9, 12), 365)
        self.assertEqual(months2days(1696, 9, -1), -31)
        self.assertEqual(months2days(1696, 9, -2), -62)
        self.assertEqual(months2days(1696, 9, -12), -366)

        self.assertEqual(months2days(1697, 2, 0), 0)
        self.assertEqual(months2days(1697, 2, 1), 28)
        self.assertEqual(months2days(1697, 2, 12), 365)
        self.assertEqual(months2days(1697, 2, -1), -31)
        self.assertEqual(months2days(1697, 2, -2), -62)
        self.assertEqual(months2days(1697, 2, -3), -92)
        self.assertEqual(months2days(1697, 2, -12), -366)
        self.assertEqual(months2days(1697, 2, -14), -428)
        self.assertEqual(months2days(1697, 2, -15), -458)

        self.assertEqual(months2days(1903, 3, 0), 0)
        self.assertEqual(months2days(1903, 3, 1), 31)
        self.assertEqual(months2days(1903, 3, 2), 61)
        self.assertEqual(months2days(1903, 3, 3), 92)
        self.assertEqual(months2days(1903, 3, 4), 122)
        self.assertEqual(months2days(1903, 3, 11), 366 - 29)
        self.assertEqual(months2days(1903, 3, 12), 366)
        self.assertEqual(months2days(1903, 3, -1), -28)
        self.assertEqual(months2days(1903, 3, -2), -59)
        self.assertEqual(months2days(1903, 3, -3), -90)
        self.assertEqual(months2days(1903, 3, -12), -365)

        self.assertEqual(months2days(1903, 7, 0), 0)
        self.assertEqual(months2days(1903, 7, 1), 31)
        self.assertEqual(months2days(1903, 7, 2), 62)
        self.assertEqual(months2days(1903, 7, 3), 92)
        self.assertEqual(months2days(1903, 7, 6), 184)
        self.assertEqual(months2days(1903, 7, 12), 366)
        self.assertEqual(months2days(1903, 7, -1), -30)
        self.assertEqual(months2days(1903, 7, -2), -61)
        self.assertEqual(months2days(1903, 7, -6), -181)
        self.assertEqual(months2days(1903, 7, -12), -365)

        # Extra tests
        self.assertEqual(months2days(1900, 3, 0), 0)
        self.assertEqual(months2days(1900, 3, 1), 31)
        self.assertEqual(months2days(1900, 3, 24), 730)
        self.assertEqual(months2days(1900, 3, -1), -28)
        self.assertEqual(months2days(1900, 3, -24), -730)

        self.assertEqual(months2days(1000, 4, 0), 0)
        self.assertEqual(months2days(1000, 4, 1), 30)
        self.assertEqual(months2days(1000, 4, 24), 730)
        self.assertEqual(months2days(1000, 4, -1), -31)
        self.assertEqual(months2days(1000, 4, -24), -730)

        self.assertEqual(months2days(2001, 10, -12), -365)
        self.assertEqual(months2days(2000, 10, -12), -366)
        self.assertEqual(months2days(2000, 2, -12), -365)
        self.assertEqual(months2days(2000, 3, -12), -366)

    def test_round_number_function(self):
        self.assertTrue(math.isnan(round_number(float('NaN'))))
        self.assertTrue(math.isinf(round_number(float('INF'))))
        self.assertTrue(math.isinf(round_number(float('-INF'))))
        self.assertEqual(round_number(10.1), 10)
        self.assertEqual(round_number(9.5), 10)
        self.assertEqual(round_number(-10.1), -10)
        self.assertEqual(round_number(-9.5), -9)

    def test_collapse_white_spaces_function(self):
        self.assertEqual(collapse_white_spaces('  ab  c  '), 'ab c')
        self.assertEqual(collapse_white_spaces('  ab\t\nc  '), 'ab c')

    def test_get_double_function(self):
        self.assertEqual(get_double(1), 1.0)
        self.assertEqual(get_double(1.0), 1.0)

        self.assertIs(get_double('NaN'), math.nan)
        self.assertIs(get_double(float('nan')), math.nan)
        self.assertTrue(math.isinf(get_double('INF')))

        self.assertRaises(ValueError, get_double, 'nan')
        self.assertRaises(ValueError, get_double, 'Inf')
        self.assertRaises(ValueError, get_double, 'alfa')

    def test_numeric_equal_function(self):
        self.assertTrue(numeric_equal(1.0, 1))
        self.assertFalse(numeric_equal(1.000001, 1.0))
        self.assertTrue(numeric_equal(1.0000001, 1.0))
        self.assertFalse(numeric_equal(float('nan'), float('nan')))
        self.assertRaises(TypeError, numeric_equal, 'xyz', 1)

    def test_numeric_not_equal_function(self):
        self.assertFalse(numeric_not_equal(1.0, 1))
        self.assertTrue(numeric_not_equal(1.000001, 1.0))
        self.assertFalse(numeric_not_equal(1.0000001, 1.0))
        self.assertTrue(numeric_not_equal(float('nan'), float('nan')))

    def test_equal_function(self):
        self.assertTrue(equal(1.0, 1))
        self.assertFalse(equal(1.000001, 1.0))
        self.assertFalse(equal(1.0000001, 1.0))
        self.assertTrue(equal(float('nan'), float('nan')))
        self.assertTrue(equal('xyz', 'xyz'))

    def test_not_equal_function(self):
        self.assertFalse(not_equal(1.0, 1))
        self.assertTrue(not_equal(1.000001, 1.0))
        self.assertTrue(not_equal(1.0000001, 1.0))
        self.assertFalse(not_equal(float('nan'), float('nan')))
        self.assertFalse(not_equal('xyz', 'xyz'))

    def test_match_wildcard_function(self):
        self.assertTrue(match_wildcard('foo', '*'))
        self.assertTrue(match_wildcard('foo', '*:*'))
        self.assertTrue(match_wildcard('foo', '*:foo'))
        self.assertFalse(match_wildcard('foo', '*:bar'))
        self.assertTrue(match_wildcard('{ns}foo', '*:foo'))
        self.assertFalse(match_wildcard('{ns}foo', '*:bar'))
        self.assertTrue(match_wildcard('tns:foo', 'tns:*'))
        self.assertFalse(match_wildcard('tns:foo', 'bar:*'))
        self.assertTrue(match_wildcard('{ns}foo', '{ns}*'))
        self.assertFalse(match_wildcard('{ns}foo', '{ns}foo'))  # is not a wildcard
        self.assertFalse(match_wildcard('{ns}foo', '{ns}bar'))

    def test_escape_json_string_function(self):
        self.assertEqual(escape_json_string("\""), '\\"')
        self.assertEqual(escape_json_string("\""), '\\"')
        self.assertEqual(escape_json_string('\\"', escaped=True), '\\"')
        self.assertEqual(escape_json_string('\\u000A', escaped=True), '\\u000A')

    def test_unescape_json_string_function(self):
        self.assertEqual(unescape_json_string('foo'), 'foo')
        self.assertEqual(unescape_json_string('\\n'), '\n')
        self.assertEqual(unescape_json_string('\\u0031'), '1')
        self.assertEqual(unescape_json_string('\\"'), '"')
        self.assertEqual(unescape_json_string('\\\\'), '\\')
        self.assertEqual(unescape_json_string('\\u000a'), '\n')
        self.assertEqual(unescape_json_string('-\\r-'), '-\r-')
        self.assertEqual(unescape_json_string("-\\t-"), '-\t-')

    def test_iter_sequence_function(self):
        self.assertListEqual(list(iter_sequence(None)), [])
        self.assertListEqual(list(iter_sequence([None, 8])), [8])
        self.assertListEqual(list(iter_sequence([])), [])
        self.assertListEqual(list(iter_sequence([[], 8])), [8])
        self.assertListEqual(list(iter_sequence([[], [], []])), [])
        self.assertListEqual(list(iter_sequence([[], 8, [9]])), [8, 9])

    def test_split_function_test_function(self):
        self.assertListEqual(
            split_function_test('element(*)'), []
        )
        self.assertListEqual(
            split_function_test('function(*)'), ['*']
        )
        self.assertListEqual(
            split_function_test('function(item()) as xs:anyAtomicType'),
            ['item()', 'xs:anyAtomicType']
        )
        self.assertListEqual(
            split_function_test('function(xs:string) as xs:integer*'),
            ['xs:string', 'xs:integer*']
        )
        self.assertListEqual(
            split_function_test('function() as map(xs:string, item())'),
            ['map(xs:string, item())']
        )
        self.assertListEqual(
            split_function_test('function(item()*, item()*, item()*) as item()*'),
            ['item()*', 'item()*', 'item()*', 'item()*']
        )


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2018-2022, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
#
# Note: Many tests are built using the examples of the XPath standards,
#       published by W3C under the W3C Document License.
#
#       References:
#           https://www.w3.org/TR/xpath-3/
#           https://www.w3.org/TR/xpath-30/
#           https://www.w3.org/TR/xpath-31/
#           https://www.w3.org/Consortium/Legal/2015/doc-license
#           https://www.w3.org/TR/charmod-norm/
#
import unittest
import os
from textwrap import dedent
from typing import cast

try:
    import lxml.etree as lxml_etree
except ImportError:
    lxml_etree = None

try:
    import xmlschema
except ImportError:
    xmlschema = None
else:
    xmlschema.XMLSchema.meta_schema.build()

from elementpath import XPathContext, select
from elementpath.etree import etree_deep_equal
from elementpath.datatypes import DateTime, Base64Binary
from elementpath.xpath_nodes import DocumentNode
from elementpath.xpath3 import XPath31Parser
from elementpath.xpath_tokens import XPathMap, XPathArray

try:
    from tests import test_xpath30
except ImportError:
    import test_xpath30

MAP_WEEKDAYS = """\
map {
  "Su" : "Sunday",
  "Mo" : "Monday",
  "Tu" : "Tuesday",
  "We" : "Wednesday",
  "Th" : "Thursday",
  "Fr" : "Friday",
  "Sa" : "Saturday"
}"""

MAP_WEEKDAYS_DE = """\
map{0:"Sonntag", 1:"Montag", 2:"Dienstag",
     3:"Mittwoch", 4:"Donnerstag", 5:"Freitag", 6:"Samstag"}"""


NESTED_MAP = """\
map {
    "book": map {
        "title": "Data on the Web",
        "year": 2000,
        "author": [
            map {
                "last": "Abiteboul",
                "first": "Serge"
            },
            map {
                "last": "Buneman",
                "first": "Peter"
            },
            map {
                "last": "Suciu",
                "first": "Dan"
            }
        ],
        "publisher": "Morgan Kaufmann Publishers",
        "price": 39.95
    }
}"""


class XPath31ParserTest(test_xpath30.XPath30ParserTest):

    def setUp(self):
        self.parser = XPath31Parser(namespaces=self.namespaces)

    def test_map_weekdays(self):
        token = self.parser.parse(MAP_WEEKDAYS)
        self.assertIsInstance(token, XPathMap)

        map_value = {'Su': 'Sunday',
                     'Mo': 'Monday',
                     'Tu': 'Tuesday',
                     'We': 'Wednesday',
                     'Th': 'Thursday',
                     'Fr': 'Friday',
                     'Sa': 'Saturday'}

        self.assertEqual(token.symbol, 'map')
        self.assertEqual(token.label, 'map')
        self.assertEqual(token.source, f'map{map_value!r}'.replace(': ', ':'))
        self.assertEqual(
            repr(token),
            f"<XPathMap object (not evaluated constructor) at {hex(id(token))}>"
        )
        self.assertEqual(str(token), 'not evaluated map constructor with 7 entries')

        self.assertDictEqual(token.evaluate()._map, map_value)

        self.assertTrue(
            repr(token.evaluate()).startswith('<XPathMap object at ')
        )
        self.assertEqual(str(token.evaluate()), f'map{map_value!r}')

        token = self.parser.parse(f"{MAP_WEEKDAYS}('Mo')")
        self.assertEqual(token.evaluate(), 'Monday')

        token = self.parser.parse(f"{MAP_WEEKDAYS}('Mon')")
        self.assertEqual(token.evaluate(), [])

        token = self.parser.parse(f"let $x := {MAP_WEEKDAYS} return $x('Mo')")
        context = XPathContext(self.etree.XML('<empty/>'))
        self.assertEqual(token.evaluate(context), ['Monday'])

    def test_nested_map(self):
        token = self.parser.parse(f'{NESTED_MAP}("book")("title")')
        self.assertEqual(token.evaluate(), 'Data on the Web')

        self.assertEqual(token.symbol, '(')
        self.assertEqual(token.label, 'expression')
        self.assertTrue(token.source.startswith("map{'book':map{'title':'Data on the Web', "))
        self.assertTrue(token.source.endswith(", 'price':39.95}}('book')('title')"))
        self.assertEqual(
            repr(token), f'<_LeftParenthesisExpression object at {hex(id(token))}>'
        )
        self.assertEqual(str(token), "function call expression")

        token = self.parser.parse(f'{NESTED_MAP}("book")("author")')
        self.assertIsInstance(token.evaluate(), XPathArray)

        token = self.parser.parse(f'{NESTED_MAP}("book")("author")(1)("last")')
        self.assertEqual(token.evaluate(), 'Abiteboul')

    def test_map_ambiguity(self):
        self.parser.namespaces['a'] = 'http://xpath.test/ns'
        try:
            with self.assertRaises(SyntaxError):
                self.parser.parse('map{a:b}')

            token = cast(XPathMap, self.parser.parse('map{a :b}'))
            self.assertEqual(token[0].symbol, '(name)')
            self.assertEqual(token[0].value, 'a')
            self.assertEqual(token._values[0].symbol, '(name)')
            self.assertEqual(token._values[0].value, 'b')

            token = cast(XPathMap, self.parser.parse('map{a: b}'))
            self.assertEqual(token[0].symbol, '(name)')
            self.assertEqual(token[0].value, 'a')
            self.assertEqual(token._values[0].symbol, '(name)')
            self.assertEqual(token._values[0].value, 'b')

            token = self.parser.parse('map{a:b:c}')
            self.assertEqual(token[0].symbol, ':')
            self.assertEqual(token[0].value, 'a:b')
            self.assertEqual(token._values[0].symbol, '(name)')
            self.assertEqual(token._values[0].value, 'c')

            token = self.parser.parse('map{a:*:c}')
            self.assertEqual(token[0].symbol, ':')
            self.assertEqual(token[0].value, 'a:*')
            self.assertEqual(token._values[0].symbol, '(name)')
            self.assertEqual(token._values[0].value, 'c')

            token = self.parser.parse('map{*:b:c}')
            self.assertEqual(token[0].symbol, ':')
            self.assertEqual(token[0].value, '*:b')
            self.assertEqual(token._values[0].symbol, '(name)')
            self.assertEqual(token._values[0].value, 'c')
        finally:
            self.parser.namespaces.pop('a')

    def test_curly_array_constructor(self):
        token = self.parser.parse('array { 1, 2, 5, 7 }')
        self.assertIsInstance(token, XPathArray)

        self.assertEqual(token.symbol, 'array')
        self.assertEqual(token.label, 'array')
        self.assertEqual(token.source, 'array{1, 2, 5, 7}')
        self.assertEqual(
            repr(token), f'<XPathArray object (not evaluated constructor) at {hex(id(token))}>'
        )
        self.assertEqual(str(token), 'not evaluated curly array constructor with 4 items')

        array = token.evaluate()  # Create a new object ...
        self.assertEqual(
            repr(token), f'<XPathArray object (not evaluated constructor) at {hex(id(token))}>'
        )
        self.assertEqual(str(array), '[1, 2, 5, 7]')

    def test_square_array_constructor(self):
        token = self.parser.parse('[ 1, 2, 5, 7 ]')
        self.assertIsInstance(token, XPathArray)

        self.assertEqual(token.symbol, '[')
        self.assertEqual(token.label, 'array')
        self.assertEqual(token.source, '[1, 2, 5, 7]')
        self.assertEqual(
            repr(token),
            f"<XPathArray object (not evaluated constructor) at {hex(id(token))}>"
        )
        self.assertEqual(str(token), 'not evaluated square array constructor with 4 items')

        array = token.evaluate()
        self.assertEqual(
            repr(array),
            f"<{array.__class__.__name__} object at {hex(id(array))}>"
        )
        self.assertEqual(str(array), '[1, 2, 5, 7]')

    def test_array_lookup(self):
        token = self.parser.parse('array { 1, 2, 5, 7 }(4)')
        self.assertEqual(token.evaluate(), 7)
        self.assertEqual(token.source, 'array{1, 2, 5, 7}(4)')
        self.assertEqual(repr(token), f'<_LeftParenthesisExpression object at {hex(id(token))}>')
        self.assertEqual(
            repr(token[0]),
            f"<XPathArray object (not evaluated constructor) at {hex(id(token[0]))}>"
        )
        self.assertEqual(str(token), "function call expression")

        token = self.parser.parse('[ 1, 2, 5, 7 ](4)')
        self.assertEqual(token.evaluate(), 7)
        self.assertEqual(token.source, '[1, 2, 5, 7](4)')
        self.assertEqual(repr(token), f'<_LeftParenthesisExpression object at {hex(id(token))}>')
        self.assertEqual(
            repr(token[0]),
            f"<XPathArray object (not evaluated constructor) at {hex(id(token[0]))}>"
        )
        self.assertEqual(str(token), "function call expression")

    def test_map_size_function(self):
        token = self.parser.parse('map:size(map{})')
        self.assertEqual(token.evaluate(), 0)
        self.assertEqual(str(token), "'map:size' function")
        self.assertEqual(
            repr(token), f'<_PrefixedReferenceToken object at {hex(id(token))}>'
        )
        self.assertEqual(token.source, 'map:size(map{})')

        self.check_value('map:size(map{"true":1, "false":0})', 2)

    def test_map_keys_function(self):
        token = self.parser.parse('map:keys(map{})')
        self.assertListEqual(token.evaluate(), [])
        self.assertEqual(str(token), "'map:keys' function")
        self.assertEqual(
            repr(token), f'<_PrefixedReferenceToken object at {hex(id(token))}>'
        )
        self.assertEqual(token.source, 'map:keys(map{})')

        self.check_value('map:keys(map{1:"yes", 2:"no"})', {1, 2})

    def test_map_contains_function(self):
        self.check_value('map:contains(map{}, 1)', False)
        self.check_value('map:contains(map{}, "xyz")', False)
        self.check_value('map:contains(map{1:"yes", 2:"no"}, 1)', True)
        self.check_value('map:contains(map{"xyz":23}, "xyz")', True)
        self.check_value('map:contains(map{"abc":23, "xyz":()}, "xyz")', True)
        self.check_source('map:contains(map{"xyz":23}, "xyz")',
                          "map:contains(map{'xyz':23}, 'xyz')")

        context = XPathContext(self.etree.XML('<empty/>'))

        expression = f"let $x := {MAP_WEEKDAYS_DE} return map:contains($x, 2)"
        self.check_value(expression, [True], context=context)

        expression = f"let $x := {MAP_WEEKDAYS_DE} return map:contains($x, 9)"
        self.check_value(expression, [False], context=context)

    def test_map_get_function(self):
        context = XPathContext(self.etree.XML('<empty/>'))

        expression = f"let $x := {MAP_WEEKDAYS} return map:get($x, 'Mo')"
        self.check_value(expression, ['Monday'], context=context)

        # Tht source property returns a compacted normalized form
        expected = expression.\
            replace('\n', '').\
            replace('\r', '').\
            replace('  ', ' ').\
            replace('"', "'").\
            replace('map {', 'map{').\
            replace('{ ', '{').replace(' : ', ':')
        self.check_source(expression, expected)

        expression = f"let $x := {MAP_WEEKDAYS} return map:get($x, 'Mon')"
        self.check_value(expression, [], context=context)

    def test_map_put_function(self):
        context = XPathContext(self.etree.XML('<empty/>'))
        expression = f'let $week := {MAP_WEEKDAYS_DE} return map:put($week, 6, "Sonnabend")'
        result = XPathMap(self.parser, items={
            0: "Sonntag", 1: "Montag", 2: "Dienstag", 3: "Mittwoch",
            4: "Donnerstag", 5: "Freitag", 6: "Sonnabend"
        })
        self.check_value(expression, [result], context=context)

        expected = expression.\
            replace('\n', '').\
            replace('\r', '').\
            replace('"', "'").\
            replace('     ', ' ')
        self.check_source(expression, expected)

    def test_map_remove_function(self):
        context = XPathContext(self.etree.XML('<empty/>'))

        expression = f'let $week := {MAP_WEEKDAYS_DE} return map:remove($week, 4)'
        result = XPathMap(self.parser, items={
            0: "Sonntag", 1: "Montag", 2: "Dienstag",
            3: "Mittwoch", 5: "Freitag", 6: "Samstag"
        })
        self.check_value(expression, [result], context=context)

        expression = f'let $week := {MAP_WEEKDAYS_DE} return map:remove($week, (0, 6 to 7))'
        result = XPathMap(self.parser, items={
            1: "Montag", 2: "Dienstag", 3: "Mittwoch", 4: "Donnerstag", 5: "Freitag"
        })
        self.check_value(expression, [result], context=context)

        expression = f'let $week := {MAP_WEEKDAYS_DE} return map:remove($week, ())'
        result = XPathMap(self.parser, items={
            0: "Sonntag", 1: "Montag", 2: "Dienstag", 3: "Mittwoch",
            4: "Donnerstag", 5: "Freitag", 6: "Samstag"
        })
        self.check_value(expression, [result], context=context)

        expected = expression.\
            replace('\n', '').\
            replace('\r', '').\
            replace('"', "'").\
            replace('     ', ' ')
        self.check_source(expression, expected)

        expression = f'let $week := {MAP_WEEKDAYS_DE} return map:remove($week, 4)'
        result = XPathMap(self.parser, items={
            0: "Sonntag", 1: "Montag", 2: "Dienstag",
            3: "Mittwoch",  # 4: "Donnerstag",
            5: "Freitag", 6: "Samstag"
        })
        self.check_value(expression, [result], context=context)

    def test_map_entry_function(self):
        context = XPathContext(self.etree.XML('<empty/>'))

        expression = 'map:entry("M", "Monday")'
        result = XPathMap(self.parser, items={'M': 'Monday'})
        self.check_value(expression, result, context=context)
        self.check_source(expression, expression.replace('"', "'"))

        # e.g.: Alternative low level token-based check
        token = self.parser.parse('map:entry("M", "Monday")')
        result = token.evaluate(context)
        self.assertIsInstance(result, XPathMap)
        self.assertEqual(len(result), 1)
        self.assertEqual(result('M', context=context), 'Monday')

    def test_map_merge_function(self):
        week = {0: "Sonntag", 1: "Montag", 2: "Dienstag", 3: "Mittwoch",
                4: "Donnerstag", 5: "Freitag", 6: "Samstag"}
        context = XPathContext(
            root=self.etree.XML('<empty/>'),
            variables={'week': XPathMap(self.parser, week)}
        )

        expression = 'map:merge(())'
        result = XPathMap(self.parser, items={})
        self.check_value(expression, result, context=context)

        expression = 'map:merge((map:entry(0, "no"), map:entry(1, "yes")))'
        result = XPathMap(self.parser, items={0: 'no', 1: 'yes'})
        self.check_value(expression, result, context=context)
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'map:merge(($week, map{7:"Unbekannt"}))'
        result = XPathMap(self.parser, items={
            0: "Sonntag", 1: "Montag", 2: "Dienstag", 3: "Mittwoch",
            4: "Donnerstag", 5: "Freitag", 6: "Samstag", 7: "Unbekannt"
        })
        self.check_value(expression, result, context=context)
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'map:merge(($week, map{6:"Sonnabend"}), map{"duplicates":"use-last"})'
        result = XPathMap(self.parser, items={
            0: "Sonntag", 1: "Montag", 2: "Dienstag", 3: "Mittwoch",
            4: "Donnerstag", 5: "Freitag", 6: "Sonnabend"
        })
        self.check_value(expression, result, context=context)
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'map:merge(($week, map{6:"Sonnabend"}), map{"duplicates":"use-first"}) '
        result = XPathMap(self.parser, items={
            0: "Sonntag", 1: "Montag", 2: "Dienstag", 3: "Mittwoch",
            4: "Donnerstag", 5: "Freitag", 6: "Samstag"
        })
        self.check_value(expression, result, context=context)
        self.check_source(expression, expression.strip().replace('"', "'"))

        expression = 'map:merge(($week, map{6:"Sonnabend"}), map{"duplicates":"combine"})'
        result = XPathMap(self.parser, items={
            0: "Sonntag", 1: "Montag", 2: "Dienstag", 3: "Mittwoch",
            4: "Donnerstag", 5: "Freitag", 6: ["Samstag", "Sonnabend"]
        })
        self.check_value(expression, result, context=context)

    def test_map_find_function(self):
        map1 = XPathMap(self.parser, {0: 'no', 1: 'yes'})
        map2 = XPathMap(self.parser, {0: 'non', 1: 'oui'})
        map3 = XPathMap(self.parser, {0: 'nein', 1: ['ja', 'doch']})

        context = XPathContext(
            root=self.etree.XML('<empty/>'),
            variables={'responses': XPathArray(self.parser, [map1, map2, map3])}
        )

        expression = 'map:find($responses, 0)'
        result = XPathArray(self.parser, items=['no', 'non', 'nein'])
        self.check_value(expression, result, context=context)

        expression = 'map:find($responses, 1)'
        result = XPathArray(self.parser, items=['yes', 'oui', ['ja', 'doch']])
        self.check_value(expression, result, context=context)

        expression = 'map:find($responses, 2)'
        result = XPathArray(self.parser, items=[])
        self.check_value(expression, result, context=context)
        self.check_source(expression, expression)

        array1 = XPathArray(self.parser, items=[])
        map1 = XPathMap(self.parser, {"name": "engine", "id": "YW678", "parts": array1})
        array2 = XPathArray(self.parser, items=[map1])
        map2 = XPathMap(self.parser, {"name": "car", "id": "QZ123", "parts": array2})

        context = XPathContext(
            root=self.etree.XML('<empty/>'),
            variables={'inventory': map2}
        )

        expression = 'map:find($inventory, "parts")'
        result = XPathArray(self.parser, items=[array2, array1])
        self.check_value(expression, result, context=context)
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'let $inventory := map{"name":"car", "id":"QZ123", ' \
                     '"parts": [map{"name":"engine", "id":"YW678", "parts":[]}]} ' \
                     'return map:find($inventory, "parts")'
        token = self.parser.parse(expression)
        self.assertEqual(token.evaluate(context), [result])

    def test_map_for_each_function(self):
        context = XPathContext(self.etree.XML('<empty/>'))

        expression = 'map:for-each(map{1:"yes", 2:"no"}, function($k, $v){$k})'
        self.check_value(expression, [1, 2], context=context)

        expression = 'distinct-values(map:for-each(map{1:"yes", 2:"no"}, ' \
                     'function($k, $v) {$v}))'
        self.check_value(expression, ['yes', 'no'], context=context)
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'map:merge(map:for-each(map{"a":1, "b":2}, ' \
                     'function($k, $v){map:entry($k, $v+1)}))'
        result = XPathMap(self.parser, {'a': 2, 'b': 3})
        self.check_value(expression, result, context=context)

    def test_array_size_function(self):
        self.check_value('array:size(["a", "b", "c"])', 3)
        self.check_value('array:size(["a", ["b", "c"]])', 2)
        self.check_value('array:size([ ])', 0)
        self.check_value('array:size([[ ]])', 1)
        self.check_source('array:size(["a", ["b", "c"]])',
                          "array:size(['a', ['b', 'c']])")

    def test_array_get_function(self):
        expression = 'array:get(["a", "b", "c"], 2)'
        self.check_value(expression, 'b')
        self.check_source(expression, expression.replace('"', "'"))

        token = self.parser.parse('array:get(["a", ["b", "c"]], 2)')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result._array, ['b', 'c'])

    def test_array_put_function(self):
        expression = ' array:put(["a", "b", "c"], 2, "d")'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ['a', 'd', 'c'])
        self.check_source(expression, expression.lstrip().replace('"', "'"))

        token = self.parser.parse('array:put(["a", "b", "c"], 2, ("d", "e"))')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ['a', ['d', 'e'], 'c'])

        token = self.parser.parse('array:put(["a"], 1, ["d", "e"])')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertIsInstance(result.items()[0], XPathArray)
        self.assertListEqual(result.items()[0].items(), ['d', 'e'])

    def test_array_insert_before_function(self):
        expression = 'array:insert-before(["a", "b", "c", "d"], 3, ("x", "y"))'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ['a', 'b', ['x', 'y'], 'c', 'd'])
        self.check_source(expression, expression.replace('"', "'"))

        token = self.parser.parse('array:insert-before(["a", "b", "c", "d"], 5, ("x", "y"))')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ['a', 'b', 'c', 'd', ['x', 'y']])

        token = self.parser.parse('array:insert-before(["a", "b", "c", "d"], 3, ["x", "y"])')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(
            result.items(), ['a', 'b', XPathArray(self.parser, ['x', 'y']), 'c', 'd']
        )

    def test_array_append_function(self):
        token = self.parser.parse('array:append(["a", "b", "c"], "d")')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ['a', 'b', 'c', 'd'])

        expression = 'array:append(["a", "b", "c"], ("d", "e"))'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ['a', 'b', 'c', ['d', 'e']])
        self.check_source(expression, expression.replace('"', "'"))

        token = self.parser.parse('array:append(["a", "b", "c"], ["d", "e"])')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(
            result.items(), ['a', 'b', 'c', XPathArray(self.parser, ['d', 'e'])]
        )

    def test_array_subarray_function(self):
        token = self.parser.parse('array:subarray(["a", "b", "c", "d"], 2)')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ['b', 'c', 'd'])

        token = self.parser.parse('array:subarray(["a", "b", "c", "d"], 5)')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [])

        token = self.parser.parse('array:subarray(["a", "b", "c", "d"], 2, 0)')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [])

        token = self.parser.parse('array:subarray(["a", "b", "c", "d"], 2, 1)')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ['b'])

        token = self.parser.parse('array:subarray(["a", "b", "c", "d"], 2, 2)')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ['b', 'c'])

        expression = 'array:subarray(["a", "b", "c", "d"], 5, 0)'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [])
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'array:subarray([ ], 1, 0)'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [])
        self.check_source(expression, expression.replace('[ ]', '[]'))

    def test_array_head_function(self):
        self.check_value('array:head([5, 6, 7, 8])', 5)
        self.check_value('array:head([("a", "b"), ("c", "d")])', ['a', 'b'])

        expression = 'array:head([["a", "b"], ["c", "d"]])'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ['a', 'b'])
        self.check_source(expression, expression.replace('"', "'"))

    def test_array_tail_function(self):
        expression = 'array:tail([5, 6, 7, 8])'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [6, 7, 8])
        self.check_source(expression, expression)

        token = self.parser.parse('array:tail([5])')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [])

    def test_array_reverse_function(self):
        token = self.parser.parse('array:reverse(["a", "b", "c", "d"])')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ["d", "c", "b", "a"])

        expression = 'array:reverse([("a", "b"), ("c", "d")])'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [["c", "d"], ["a", "b"]])
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'array:reverse([(1 to 5)])'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [[1, 2, 3, 4, 5]])
        self.check_source(expression, expression)

        token = self.parser.parse('array:reverse([])')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [])

    def test_array_remove_function(self):
        token = self.parser.parse('array:remove(["a", "b", "c", "d"], 1)')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ["b", "c", "d"])

        token = self.parser.parse('array:remove(["a", "b", "c", "d"], 2)')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ["a", "c", "d"])

        token = self.parser.parse('array:remove(["a"], 1)')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [])

        expression = 'array:remove(["a", "b", "c", "d"], 1 to 3)'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ["d"])
        self.check_source(expression, expression.replace('"', "'"))

        token = self.parser.parse('array:remove(["a", "b", "c", "d"], ())')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ["a", "b", "c", "d"])

        self.wrong_value('array:remove(["a", "b", "c", "d"], 0)', 'FOAY0001')

    def test_array_join_function(self):
        token = self.parser.parse('array:join(())')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [])

        token = self.parser.parse('array:join([1, 2, 3])')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [1, 2, 3])

        token = self.parser.parse(' array:join((["a", "b"], ["c", "d"]))')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ["a", "b", "c", "d"])

        token = self.parser.parse('array:join((["a", "b"], ["c", "d"], [ ]))')
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ["a", "b", "c", "d"])

        expression = 'array:join((["a", "b"], ["c", "d"], [["e", "f"]]))'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(
            result.items(), ["a", "b", "c", "d", XPathArray(self.parser, ['e', 'f'])]
        )
        self.check_source(expression, expression.replace('"', "'"))

    def test_array_flatten_function(self):
        token = self.parser.parse('array:flatten([1, 4, 6, 5, 3])')
        result = token.evaluate()
        self.assertListEqual(result, [1, 4, 6, 5, 3])

        expression = 'array:flatten(([1, 2, 5], [[10, 11], 12], [], 13))'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertListEqual(result, [1, 2, 5, 10, 11, 12, 13])
        self.check_source(expression, expression)

        expression = 'array:flatten([(1, 0), (1, 1), (0, 1), (0, 0)])'
        token = self.parser.parse(expression)
        result = token.evaluate()
        self.assertListEqual(result, [1, 0, 1, 1, 0, 1, 0, 0])
        self.check_source(expression, expression)

    def test_array_for_each_function(self):
        context = XPathContext(self.etree.XML('<empty/>'))

        expression = 'array:for-each(["A", "B", 1, 2], function($z) {$z instance of xs:integer})'
        token = self.parser.parse(expression)
        result = token.evaluate(context)
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [False, False, True, True])

        expression = 'array:for-each(["the cat", "sat", "on the mat"], fn:tokenize#1)'
        token = self.parser.parse(expression)
        result = token.evaluate(context)
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [["the", "cat"], "sat", ["on", "the", "mat"]])
        self.check_source(expression, expression.replace('"', "'"))

    def test_array_for_each_pair_function(self):
        context = XPathContext(self.etree.XML('<empty/>'))

        expression = 'array:for-each-pair(["A", "B", "C"], [1, 2, 3], ' \
                     'function($x, $y) { array {$x, $y}})'
        token = self.parser.parse(expression)
        result = token.evaluate(context)
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [
            XPathArray(self.parser, ['A', 1]),
            XPathArray(self.parser, ['B', 2]),
            XPathArray(self.parser, ['C', 3])
        ])
        expected = expression.replace('"', "'").replace('{ array ', '{array')
        self.check_source(expression, expected)

        expression = 'let $A := ["A", "B", "C", "D"] ' \
                     'return array:for-each-pair($A, array:tail($A), concat#2)'
        token = self.parser.parse(expression)
        result = token.evaluate(context)
        self.assertListEqual(result, [XPathArray(self.parser, ['AB', 'BC', 'CD'])])
        self.check_source(expression, expression.replace('"', "'"))

    def test_array_filter_function(self):
        context = XPathContext(self.etree.XML('<empty/>'))

        expression = 'array:filter(["A", "B", 1, 2], function($x) {$x instance of xs:integer})'
        token = self.parser.parse(expression)
        result = token.evaluate(context)
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), [1, 2])

        expression = 'array:filter(["the cat", "sat", "on the mat"], ' \
                     'function($s) {fn:count(fn:tokenize($s)) gt 1})'
        token = self.parser.parse(expression)
        result = token.evaluate(context)
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ["the cat", "on the mat"])
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'array:filter(["A", "B", "", 0, 1], boolean#1)'
        token = self.parser.parse(expression)
        result = token.evaluate(context)
        self.assertIsInstance(result, XPathArray)
        self.assertListEqual(result.items(), ["A", "B", 1])
        self.check_source(expression, expression.replace('"', "'"))

    def test_array_fold_left_function(self):
        context = XPathContext(self.etree.XML('<empty/>'))

        expression = 'array:fold-left([true(), true(), false()], true(), ' \
                     'function($x, $y){$x and $y})'
        self.check_value(expression, [False], context=context)

        expression = 'array:fold-left([true(), true(), false()], false(), ' \
                     'function($x, $y){$x or $y})'
        self.check_value(expression, [True], context=context)

        expression = 'array:fold-left([1, 2, 3], [], function($x, $y){[$x, $y]})'
        ar1 = XPathArray(self.parser, [])
        ar2 = XPathArray(self.parser, items=[ar1, 1])
        ar3 = XPathArray(self.parser, items=[ar2, 2])
        ar4 = XPathArray(self.parser, items=[ar3, 3])
        self.check_value(expression, [ar4], context=context)
        self.check_source(expression, expression.replace('){', ') {'))

    def test_array_fold_right_function(self):
        context = XPathContext(self.etree.XML('<empty/>'))

        expression = 'array:fold-right([true(), true(), false()], true(), ' \
                     'function($x, $y){$x and $y})'
        self.check_value(expression, [False], context=context)

        expression = 'array:fold-right([true(), true(), false()], false(), ' \
                     'function($x, $y){$x or $y})'
        self.check_value(expression, [True], context=context)

        expression = 'array:fold-right([1,2,3], [], function($x, $y){[$x, $y]})'
        ar1 = XPathArray(self.parser, [])
        ar2 = XPathArray(self.parser, items=[3, ar1])
        ar3 = XPathArray(self.parser, items=[2, ar2])
        ar4 = XPathArray(self.parser, items=[1, ar3])
        self.check_value(expression, [ar4], context=context)
        self.check_source(expression, expression.replace('){', ') {').replace('2,', ' 2, '))

    def test_array_sort_function(self):
        expression = 'array:sort([1, 4, 6, 5, 3])'
        self.check_value(expression, XPathArray(self.parser, [1, 3, 4, 5, 6]))

        expression = 'array:sort([1, -2, 5, 10, -10, 10, 8], (), fn:abs#1)'
        self.check_value(expression, XPathArray(self.parser, [1, -2, 5, 8, 10, -10, 10]))
        self.check_source(expression, expression)

        expression = 'array:sort([(1,0), (1,1), (0,1), (0,0)])'
        self.check_value(expression, XPathArray(self.parser, [[0, 0], [0, 1], [1, 0], [1, 1]]))

    def test_sort_function(self):
        expression = 'fn:sort((1, 4, 6, 5, 3))'
        self.check_value(expression, [1, 3, 4, 5, 6])

        expression = 'fn:sort((1, -2, 5, 10, -10, 10, 8), (), fn:abs#1)'
        self.check_value(expression, [1, -2, 5, 8, 10, -10, 10])
        self.check_source(expression, expression)

    def test_parse_json_function(self):
        expression = 'parse-json(\'{"x":1, "y":[3,4,5]}\')'
        result = XPathMap(self.parser, {'x': 1, 'y': XPathArray(self.parser, [3, 4, 5])})
        self.check_value(expression, result)

        expression = 'parse-json(\'"abcd"\')'
        self.check_value(expression, 'abcd')

        expression = 'parse-json(\'{"x":"\\\\", "y":"\\u0025"}\')'
        result = XPathMap(self.parser, {"x": "\\", "y": "%"})
        self.check_value(expression, result)
        self.check_source(expression, expression)

        expression = 'parse-json(\'{"x":"\\\\", "y":"\\u0025"}\', map{\'escape\':true()})'
        result = XPathMap(self.parser, {"x": "\\\\", "y": "%"})
        self.check_value(expression, result)

        expression = 'parse-json(\'{"x":"\\\\", "y":"\\u0000"}\', ' \
                     'map{\'fallback\':function($s){\'[\'||$s||\']\'}})'
        result = XPathMap(self.parser, {"x": "\\", "y": "[\\u0000]"})

        # fallback inline function requires a context for evaluation
        context = XPathContext(root=self.etree.XML('<empty/>'))
        self.check_value(expression, result, context=context)

    def test_load_xquery_module_function(self):
        self.wrong_value('load-xquery-module("")', 'FOQM0001')

        with self.assertRaises(RuntimeError) as ctx:
            self.check_value('load-xquery-module("./xquery-module")')

        self.assertIn('FOQM0006', str(ctx.exception))

    def test_transform_function(self):
        with self.assertRaises(RuntimeError) as ctx:
            self.check_value('transform(map{})')

        self.assertIn('FOXT0004', str(ctx.exception))

    def test_random_number_generator_function(self):
        context = None

        expression = 'random-number-generator()'
        token = self.parser.parse(expression)
        self.assertEqual(token.source, expression)
        result = token.evaluate()
        self.assertIsInstance(result, XPathMap)

        self.assertListEqual(list(result.keys()), ['number', 'next', 'permute'])
        self.assertTrue(0 <= result('number', context=context) <= 1)

        seq = result('permute', context=context)(range(10))
        _seq = tuple(seq)
        self.assertNotEqual(seq, list(range(10)))
        self.assertNotEqual(seq, result('permute', context=context)(seq))
        self.assertNotEqual(seq, result('permute', context=context)(range(10)))
        self.assertListEqual(seq, list(_seq))

        expression = 'random-number-generator(1000)'
        token = self.parser.parse(expression)
        self.assertEqual(token.source, expression)
        result = token.evaluate()
        self.assertNotEqual(seq, result('permute', context=context)(seq))

    def test_apply_function(self):
        expression = 'fn:apply(fn:concat#3, ["a", "b", "c"])'
        self.check_value(expression, 'abc')
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'fn:apply(fn:concat#3, ["a", "b", "c", "d"])'
        self.wrong_type(expression, 'FOAP0001')

        expression = 'fn:apply(fn:concat#4, array:subarray(["a", "b", "c", "d", "e", "f"], ' \
                     '1, fn:function-arity(fn:concat#4)))'
        self.check_value(expression, 'abcd')
        self.check_source(expression, expression.replace('"', "'"))

    def test_parse_ietf_date_function(self):
        expression = 'fn:parse-ietf-date("Wed, 06 Jun 1994 07:29:35 GMT")'
        result = DateTime.fromstring('1994-06-06T07:29:35Z')
        self.check_value(expression, result)
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'fn:parse-ietf-date("Wed, 6 Jun 94 07:29:35 GMT")'
        result = DateTime.fromstring('1994-06-06T07:29:35Z')
        self.check_value(expression, result)
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'fn:parse-ietf-date("Wed Jun 06 11:54:45 EST 2013")'
        result = DateTime.fromstring('2013-06-06T11:54:45-05:00')
        self.check_value(expression, result)
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'fn:parse-ietf-date("Sunday, 06-Nov-94 08:49:37 GMT")'
        result = DateTime.fromstring('1994-11-06T08:49:37Z')
        self.check_value(expression, result)
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'fn:parse-ietf-date("Wed, 6 Jun 94 07:29:35 +0500")'
        result = DateTime.fromstring('1994-06-06T07:29:35+05:00')
        self.check_value(expression, result)
        self.check_source(expression, expression.replace('"', "'"))

    def test_contains_token_function(self):
        expression = 'fn:contains-token("red green blue ", "red")'
        self.check_value(expression, True)

        expression = 'fn:contains-token(("red", "green", "blue"), " red ")'
        self.check_value(expression, True)
        self.check_source(expression, expression.replace('"', "'"))

        expression = 'fn:contains-token("red, green, blue", "red")'
        self.check_value(expression, False)

        expression = \
            'fn:contains-token("red green blue", "RED", ' \
            '"http://www.w3.org/2005/xpath-functions/collation/html-ascii-case-insensitive")'
        self.check_value(expression, True)
        self.check_source(expression, expression.replace('"', "'"))

    def test_collation_key_function(self):
        expression = 'fn:collation-key("foo")'
        self.check_value(expression, Base64Binary(b'Zm9v'))
        self.check_source(expression, expression.replace('"', "'"))

    def test_lookup_unary_operator(self):
        context = XPathContext(self.etree.XML('<empty/>'))

        expression = '([1, 2, 3], [1, 2, 5], [1, 2, 6])[?3 = 5]'
        result = [XPathArray(self.parser, [1, 2, 5])]
        self.check_value(expression, result, context=context)
        self.check_source(expression, expression)

    def test_lookup_postfix_operator(self):
        expression = '[1, 2, 5, 7]?*'
        self.check_value(expression, [1, 2, 5, 7])
        self.check_source(expression, expression)

        expression = '[[1, 2, 3], [4, 5, 6]]?*'
        result = [
            XPathArray(self.parser, [1, 2, 3]),
            XPathArray(self.parser, [4, 5, 6])
        ]
        self.check_value(expression, result)
        self.check_source(expression, expression)

        expression = 'map { "first" : "Jenna", "last" : "Scott" }?first'
        self.check_value(expression, ['Jenna'])

        self.check_value('[4, 5, 6]?2', [5])

        expression = '(map {"first": "Tom"}, map {"first": "Dick"}, ' \
                     'map {"first": "Harry"})?first'
        self.check_value(expression, ['Tom', 'Dick', 'Harry'])
        expected = expression.\
            replace('"', "'").\
            replace('map ', 'map').\
            replace(': ', ':')
        self.check_source(expression, expected)

        expression = '([1,2,3], [4,5,6])?2'
        self.check_value(expression, [2, 5])
        self.check_source(expression, '([1, 2, 3], [4, 5, 6])?2')

        self.wrong_value('["a","b"]?3', 'FOAY0001')

    def test_lookup_operator_tree(self):
        self.check_tree('$a?2?1', '(? (? ($ (a)) (2)) (1))')
        self.check_tree('$a?2 and $a?3', '(and (? ($ (a)) (2)) (? ($ (a)) (3)))')

        self.check_tree('$a?2?1 and $a?3?4',
                        '(and (? (? ($ (a)) (2)) (1)) (? (? ($ (a)) (3)) (4)))')
        self.check_tree('$a[1] eq 1 and $a[2] eq 2',
                        '(and (eq ([ ($ (a)) (1)) (1)) (eq ([ ($ (a)) (2)) (2)))')
        self.check_tree(
            '$a[1]?2 eq 1 and $a[2]?2 eq 2',
            '(and (eq (? ([ ($ (a)) (1)) (2)) (1)) (eq (? ([ ($ (a)) (2)) (2)) (2)))'
        )

    def test_arrow_operator(self):
        expression = '"foo" => $f("bar")'
        self.check_tree(expression, "(=> ('foo') ($ (f)) ('bar'))")
        self.check_source(expression, expression.replace('"', "'"))

        expression = '"foo" => $f()'
        self.check_tree(expression, "(=> ('foo') ($ (f)) ())")

        expression = '"foo" => upper-case()'
        # self.check_tree(expression, "(=> ('foo') (upper-case) ())")
        self.check_value(expression, 'FOO')
        self.check_source(expression, expression.replace('"', "'"))

    def test_xml_to_json_function(self):
        root = self.etree.XML('<array xmlns="http://www.w3.org/2005/xpath-functions">'
                              '<number>1</number><string>is</string><boolean>1</boolean>'
                              '</array>')

        expression = 'fn:xml-to-json(.)'
        context = XPathContext(root)
        result = '[1,"is",true]'
        self.check_value(expression, result, context=context)
        self.check_source(expression, expression)

        root = self.etree.XML('<map xmlns="http://www.w3.org/2005/xpath-functions">'
                              '<number key="Sunday">1</number><number key="Monday">2</number>'
                              '</map>')

        context = XPathContext(root)
        result = '{"Sunday":1,"Monday":2}'
        self.check_value(expression, result, context=context)

    def test_json_to_xml_function(self):
        context = XPathContext(root=self.etree.XML('<empty/>'))
        root = self.etree.XML(dedent("""\
            <map xmlns="http://www.w3.org/2005/xpath-functions">
              <number key="x">1</number>
              <array key="y">
                <number>3</number>
                <number>4</number>
                <number>5</number>
              </array>
            </map>"""))

        expression = 'json-to-xml(\'{"x": 1, "y": [3,4,5]}\')'
        token = self.parser.parse(expression)
        self.check_source(expression, expression)

        result = token.evaluate(context)
        self.assertIsInstance(result, DocumentNode)
        self.assertTrue(etree_deep_equal(result.value.getroot(), root))

        root = self.etree.XML(dedent("""\
             <string xmlns="http://www.w3.org/2005/xpath-functions">abcd</string>"""))

        token = self.parser.parse('json-to-xml(\'"abcd"\', map{\'liberal\': false()})')
        result = token.evaluate(context)
        self.assertIsInstance(result, DocumentNode)
        self.assertTrue(etree_deep_equal(result.value.getroot(), root))

        root = self.etree.XML(dedent("""\
            <map xmlns="http://www.w3.org/2005/xpath-functions">
              <string key="x">\\</string>
              <string key="y">%</string>
            </map>"""))

        expression = 'json-to-xml(\'{"x": "\\\\", "y": "\\u0025"}\')'
        token = self.parser.parse(expression)
        self.check_source(expression, expression)

        result = token.evaluate(context)
        self.assertIsInstance(result, DocumentNode)
        self.assertTrue(etree_deep_equal(result.value.getroot(), root))

        root = self.etree.XML(dedent("""\
            <map xmlns="http://www.w3.org/2005/xpath-functions">
              <string escaped="true" key="x">\\\\</string>
              <string key="y">%</string>
            </map>"""))

        expression = 'json-to-xml(\'{"x": "\\\\", "y": "\\u0025"}\', ' \
                     'map{\'escape\':true()})'
        token = self.parser.parse(expression)
        self.check_source(expression, expression)

        result = token.evaluate(context)
        self.assertIsInstance(result, DocumentNode)
        self.assertTrue(etree_deep_equal(result.value.getroot(), root))


@unittest.skipIf(lxml_etree is None, "The lxml library is not installed")
class LxmlXPath31ParserTest(XPath31ParserTest):
    etree = lxml_etree

    def test_regression_ep415_ep420__issue_71(self):
        import lxml.html as lxml_html

        xml_source = dedent("""\
        <hotel>
          <branch location="California">
            <staff>
              <date>2023-10-10</date>
              <given_name>Christopher</given_name>
              <surname>Anderson</surname>
              <age>25</age>
            </staff>
            <staff>
              <date>2023-10-11</date>
              <given_name>Christopher</given_name>
              <surname>Carter</surname>
              <age>30</age>
            </staff>
          </branch>
          <branch location="Las Vegas">
            <staff>
              <given_name>Lisa</given_name>
              <surname>Walker</surname>
              <age>60</age>
            </staff>
            <staff>
              <given_name>Jessica</given_name>
              <surname>Walker</surname>
              <age>32</age>
            </staff>
            <staff>
              <given_name>Jennifer</given_name>
              <surname>Roberts</surname>
              <age>50</age>
            </staff>
          </branch>
        </hotel>
        """)

        queries = [
            'if (count(//hotel/branch/staff) = 5) then true() else false()',
            '//hotel/branch/staff',
            'if (count(/hotel/branch/staff) = 5) then true() else false()',
            '(count(/hotel/branch/staff) = 5)',
            '(count(/hotel/branch/staff))',
            '/hotel/branch/staff',
            'for $i in /hotel/branch/staff return $i/given_name',
            'for $i in //hotel/branch/staff return $i/given_name',
            'distinct-values(for $i in /hotel/branch/staff return $i/given_name)',
            'distinct-values(for $i in //hotel/branch/staff return $i/given_name)',
            'date(/hotel/branch[1]/staff[1]/date) instance of xs:date',
            '/hotel/branch[1]/staff[1]/date cast as xs:date',
        ]

        html_parser = lxml_html.HTMLParser()
        xml_parser = lxml_etree.XMLParser(strip_cdata=False)

        xml_data = bytes(xml_source, encoding='utf-8')
        data_trees = {
            'html': lxml_html.fromstring(xml_data, parser=html_parser),
            'xml': lxml_etree.fromstring(xml_data, parser=xml_parser)
        }

        for query in queries:
            results = []
            for doctype, document in data_trees.items():
                try:
                    res = select(document, query, parser=XPath31Parser)
                except Exception as e:
                    results.append(e)
                else:
                    results.append(res)

            if isinstance(results[0], list):
                self.assertIsInstance(results[1], list)
                self.assertEqual(len(results[0]), len(results[1]))
                for e1, e2 in zip(*results):
                    self.assertEqual(getattr(e1, 'tag', e1), getattr(e2, 'tag', e2))
            else:
                self.assertEqual(results[0], results[1])


class XPath31FunctionsTest(test_xpath30.XPath30FunctionsTest):

    maxDiff = 1024

    def setUp(self):
        self.parser = XPath31Parser(namespaces=self.namespaces)

        # Make sure the tests are repeatable.
        env_vars_to_tweak = 'LC_ALL', 'LANG'
        self.current_env_vars = {v: os.environ.get(v) for v in env_vars_to_tweak}
        for v in self.current_env_vars:
            os.environ[v] = 'en_US.UTF-8'

    def tearDown(self):
        if hasattr(self, 'current_env_vars'):
            for v in self.current_env_vars:
                if self.current_env_vars[v] is not None:
                    os.environ[v] = self.current_env_vars[v]


@unittest.skipIf(lxml_etree is None, "The lxml library is not installed")
class LxmlXPath31FunctionsTest(XPath31FunctionsTest):
    etree = lxml_etree


class XPath31ConstructorsTest(test_xpath30.XPath30ConstructorsTest):
    def setUp(self):
        self.parser = XPath31Parser(namespaces=self.namespaces)


@unittest.skipIf(lxml_etree is None, "The lxml library is not installed")
class LxmlXPath31ConstructorsTest(XPath31ConstructorsTest):
    etree = lxml_etree


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2018-2021, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
from unittest.mock import patch
from textwrap import dedent
import io
import xml.etree.ElementTree as ElementTree

try:
    import lxml.etree as lxml_etree
except ImportError:
    lxml_etree = None

try:
    import xmlschema
except ImportError:
    xmlschema = None
else:
    xmlschema.XMLSchema.meta_schema.build()

from elementpath.etree import is_etree_element, etree_iter_strings, \
    etree_deep_equal, etree_iter_paths
from elementpath.xpath_nodes import DocumentNode, ElementNode, AttributeNode, TextNode, \
    NamespaceNode, CommentNode, ProcessingInstructionNode
from elementpath.tree_builders import get_node_tree
from elementpath.xpath_context import XPathContext, XPathSchemaContext


class DummyXsdType:
    name = local_name = None

    def is_matching(self, name, default_namespace): pass
    def is_empty(self): pass
    def is_simple(self): pass
    def has_simple_content(self): pass
    def has_mixed_content(self): pass
    def is_element_only(self): pass
    def is_key(self): pass
    def is_qname(self): pass
    def is_notation(self): pass
    def decode(self, obj, *args, **kwargs): return int(obj)
    def validate(self, obj, *args, **kwargs): pass


class XPathNodesTest(unittest.TestCase):
    elem = ElementTree.XML('<node a1="10"/>')

    def setUp(self):
        root = ElementTree.Element('root')
        self.context = XPathContext(root)  # Dummy context for creating nodes

    def test_is_etree_element_function(self):
        self.assertTrue(is_etree_element(self.elem))
        self.assertFalse(is_etree_element('text'))
        self.assertFalse(is_etree_element(None))

    def test_elem_iter_strings_function(self):
        root = ElementTree.XML('<A>text1\n<B1>text2</B1>tail1<B2/><B3><C1>text3</C1></B3>tail2</A>')
        result = ['text1\n', 'text2', 'tail1', 'tail2', 'text3']
        self.assertListEqual(list(etree_iter_strings(root)), result)

        with patch.multiple(DummyXsdType, has_mixed_content=lambda x: True):
            xsd_type = DummyXsdType()
            typed_root = ElementNode(elem=root, xsd_type=xsd_type)
            self.assertListEqual(list(etree_iter_strings(typed_root.elem)), result)

        norm_result = ['text1', 'text2', 'tail1', 'tail2', 'text3']
        with patch.multiple(DummyXsdType, is_element_only=lambda x: True):
            xsd_type = DummyXsdType()
            typed_root = ElementNode(elem=root, xsd_type=xsd_type)
            self.assertListEqual(list(etree_iter_strings(typed_root.elem, True)), norm_result)

            comment = ElementTree.Comment('foo')
            root[1].append(comment)
            self.assertListEqual(list(etree_iter_strings(typed_root.elem, True)), norm_result)

        self.assertListEqual(list(etree_iter_strings(root)), result)

    def test_etree_deep_equal_function(self):
        root = ElementTree.XML('<A><B1>10</B1><B2 max="20"/>end</A>')
        self.assertTrue(etree_deep_equal(root, root))

        elem = ElementTree.XML('<A><B1>11</B1><B2 max="20"/>end</A>')
        self.assertFalse(etree_deep_equal(root, elem))

        elem = ElementTree.XML('<A><B1>10</B1>30<B2 max="20"/>end</A>')
        self.assertFalse(etree_deep_equal(root, elem))

        elem = ElementTree.XML('<A xmlns:ns="tns"><B1>10</B1><B2 max="20"/>end</A>')
        self.assertTrue(etree_deep_equal(root, elem))

        elem = ElementTree.XML('<A><B1>10</B1><B2 max="20"><C1/></B2>end</A>')
        self.assertFalse(etree_deep_equal(root, elem))

    def test_match_name_method(self):
        attr = AttributeNode('a1', '10', parent=None)
        self.assertTrue(attr.match_name('*'))
        self.assertTrue(attr.match_name('a1'))
        self.assertTrue(attr.match_name('*:a1'))
        self.assertFalse(attr.match_name('{foo}*'))
        self.assertFalse(attr.match_name('foo:*'))

        self.assertTrue(
            AttributeNode('{foo}a1', '10').match_name('{foo}*')
        )

        attr = AttributeNode('{http://xpath.test/ns}a1', '10', parent=None)
        self.assertTrue(attr.match_name('*:a1'))

    def test_node_base_uri(self):
        xml_test = '<A xmlns:xml="http://www.w3.org/XML/1998/namespace" xml:base="/" />'

        self.assertEqual(ElementNode(ElementTree.XML(xml_test)).base_uri, '/')
        document = ElementTree.parse(io.StringIO(xml_test))

        self.assertIsNone(DocumentNode(document).base_uri, '/')
        self.assertIsNone(ElementNode(self.elem).base_uri)
        self.assertIsNone(TextNode('a text node').base_uri)

        xml_test = dedent("""\
            <?xml version="1.0"?>
            <e1 xml:base="http://example.org/wine/">
              <e2 xml:base="rosé"/>
            </e1>""")

        root_node = get_node_tree(ElementTree.XML(xml_test))
        self.assertEqual(root_node.base_uri, 'http://example.org/wine/')
        self.assertIsInstance(root_node[0], TextNode)
        self.assertEqual(root_node[1].base_uri, 'http://example.org/wine/rosé')

        xml_test = dedent("""\
            <collection xml:base="http://example.test/xpath/ ">
              <item xml:base="urn:isbn:0451450523"/>
              <item xml:base="urn:isan:0000-0000-2CEA-0000-1-0000-0000-Y "/>
              <item xml:base="urn:ISSN:0167-6423 "/>
              <item xml:base=" urn:ietf:rfc:2648 "/>
              <item xml:base="urn:uuid:6e8bc430-9c3a-11d9-9669-0800200c9a66 "/>
            </collection>""")

        root_node = get_node_tree(ElementTree.XML(xml_test))
        self.assertEqual(root_node.base_uri, 'http://example.test/xpath/')
        self.assertIsInstance(root_node[0], TextNode)
        self.assertEqual(root_node[0].base_uri, 'http://example.test/xpath/')
        self.assertEqual(root_node[1].base_uri, 'urn:isbn:0451450523')
        self.assertIsInstance(root_node[2], TextNode)
        self.assertEqual(root_node[3].base_uri, 'urn:isan:0000-0000-2CEA-0000-1-0000-0000-Y')
        self.assertEqual(root_node[5].base_uri, 'urn:ISSN:0167-6423')
        self.assertEqual(root_node[7].base_uri, 'urn:ietf:rfc:2648')
        self.assertEqual(root_node[9].base_uri, 'urn:uuid:6e8bc430-9c3a-11d9-9669-0800200c9a66')

    def test_node_document_uri_function(self):
        node = ElementNode(self.elem)
        self.assertIsNone(node.document_uri)

        xml_test = '<A xmlns:xml="http://www.w3.org/XML/1998/namespace" xml:base="/root" />'
        document = ElementTree.parse(io.StringIO(xml_test))
        node = DocumentNode(document)
        self.assertIsNone(node.document_uri)

        node = DocumentNode(document, uri=' http://xpath.test/doc.xml ')
        self.assertEqual(node.document_uri, 'http://xpath.test/doc.xml')

        xml_test = '<A xmlns:xml="http://www.w3.org/XML/1998/namespace" ' \
                   'xml:base="http://xpath.test" />'
        document = ElementTree.parse(io.StringIO(xml_test))
        node = DocumentNode(document)
        self.assertIsNone(node.document_uri)

        xml_test = '<A xmlns:xml="http://www.w3.org/XML/1998/namespace" xml:base="dir1/dir2" />'
        document = ElementTree.parse(io.StringIO(xml_test))
        node = DocumentNode(document, uri="dir1/dir2")
        self.assertIsNone(node.document_uri)

        xml_test = '<A xmlns:xml="http://www.w3.org/XML/1998/namespace" ' \
                   'xml:base="http://[xpath.test" />'
        document = ElementTree.parse(io.StringIO(xml_test))
        node = DocumentNode(document, uri="http://[xpath.test")
        self.assertIsNone(node.document_uri)

    def test_attribute_nodes(self):
        parent = self.context.root
        attribute = AttributeNode('id', '0212349350')

        self.assertEqual(repr(attribute),
                         "AttributeNode(name='id', value='0212349350')")
        self.assertNotEqual(attribute, AttributeNode('id', '0212349350'))
        self.assertEqual(attribute.as_item(), ('id', '0212349350'))
        self.assertNotEqual(attribute.as_item(),
                            AttributeNode('id', '0212349350'))
        self.assertNotEqual(attribute, AttributeNode('id', '0212349350', parent))

        attribute = AttributeNode('id', '0212349350', parent)
        self.assertNotEqual(attribute, AttributeNode('id', '0212349350', parent))
        self.assertEqual(attribute.as_item(), ('id', '0212349350'))

        attribute = AttributeNode('value', '10', parent)
        self.assertEqual(repr(attribute), "AttributeNode(name='value', value='10')")

        with patch.multiple(DummyXsdType, is_simple=lambda x: True):
            xsd_type = DummyXsdType()
            attribute.xsd_type = xsd_type
            self.assertEqual(attribute.as_item(), ('value', '10'))

    def test_typed_element_nodes(self):
        element = ElementTree.Element('schema')

        with patch.multiple(DummyXsdType, is_simple=lambda x: True):
            xsd_type = DummyXsdType()
            context = XPathContext(element)
            context.root.xsd_type = xsd_type
            self.assertTrue(repr(context.root).startswith(
                "ElementNode(elem=<Element 'schema' at 0x"
            ))

    def test_text_nodes(self):
        parent = self.context.root

        # equality if and only is the same instance
        text_node = TextNode('alpha')
        self.assertEqual(text_node, text_node)
        self.assertNotEqual(text_node, TextNode('alpha'))

        text_node = TextNode('alpha', parent)
        self.assertEqual(text_node, text_node)
        self.assertNotEqual(text_node, TextNode('alpha', parent))

        text_node = TextNode('alpha', parent)
        self.assertEqual(text_node, text_node)
        self.assertNotEqual(text_node, TextNode('alpha', parent))

        self.assertEqual(repr(TextNode('alpha')), "TextNode(value='alpha')")
        text_node = TextNode('alpha', parent)
        self.assertEqual(repr(text_node), "TextNode(value='alpha')")

    def test_namespace_nodes(self):
        context = self.context
        namespace = NamespaceNode('tns', 'http://xpath.test/ns')

        self.assertEqual(repr(namespace),
                         "NamespaceNode(prefix='tns', uri='http://xpath.test/ns')")
        self.assertEqual(namespace.value, 'http://xpath.test/ns')
        self.assertNotEqual(namespace, NamespaceNode('tns', 'http://xpath.test/ns'))
        self.assertEqual(namespace.as_item(), ('tns', 'http://xpath.test/ns'))
        self.assertNotEqual(
            namespace, NamespaceNode('tns', 'http://xpath.test/ns', parent=context.root)
        )

        namespace = NamespaceNode('tns', 'http://xpath.test/ns', parent=context.root)
        self.assertEqual(repr(namespace), "NamespaceNode(prefix='tns', uri='http://xpath.test/ns')")

        self.assertNotEqual(namespace,
                            NamespaceNode('tns', 'http://xpath.test/ns', parent=context.root))
        self.assertEqual(namespace.as_item(), ('tns', 'http://xpath.test/ns'))
        self.assertNotEqual(namespace, NamespaceNode('tns', 'http://xpath.test/ns'))

    def test_node_children_function(self):
        self.assertListEqual(ElementNode(self.elem).children, [])
        elem = ElementNode(ElementTree.XML("<A><B1/><B2/></A>"))
        self.assertListEqual(elem.children, [x for x in elem])

        document = DocumentNode(ElementTree.parse(io.StringIO("<A><B1/><B2/></A>")))
        self.assertListEqual(document.children, [])  # not built document
        document.children.append(ElementNode(document.value.getroot(), document))
        self.assertListEqual(document.children, [document.getroot()])

        self.assertIsNone(TextNode('a text node').children)

    def test_node_nilled_property(self):
        xml_test = '<A xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:nil="true" />'
        self.assertTrue(ElementNode(ElementTree.XML(xml_test)).nilled)
        xml_test = '<A xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:nil="false" />'
        self.assertFalse(ElementNode(ElementTree.XML(xml_test)).nilled)
        self.assertFalse(ElementNode(ElementTree.XML('<A />')).nilled)
        self.assertFalse(TextNode('foo').nilled)

    def test_node_kind_property(self):
        document = DocumentNode(ElementTree.parse(io.StringIO(u'<A/>')))
        element = ElementNode(ElementTree.Element('schema'))
        attribute = AttributeNode('id', '0212349350')
        namespace = NamespaceNode('xs', 'http://www.w3.org/2001/XMLSchema')
        comment = CommentNode(ElementTree.Comment('nothing important'))
        pi = ProcessingInstructionNode(
            self.context, ElementTree.ProcessingInstruction('action', 'nothing to do')
        )
        text = TextNode('betelgeuse')

        self.assertEqual(document.kind, 'document')
        self.assertEqual(element.kind, 'element')
        self.assertEqual(attribute.kind, 'attribute')
        self.assertEqual(namespace.kind, 'namespace')
        self.assertEqual(comment.kind, 'comment')
        self.assertEqual(pi.kind, 'processing-instruction')
        self.assertEqual(text.kind, 'text')

        with patch.multiple(DummyXsdType, is_simple=lambda x: True):
            xsd_type = DummyXsdType()

            attribute = AttributeNode('id', '0212349350', xsd_type=xsd_type)
            self.assertEqual(attribute.kind, 'attribute')

            typed_element = ElementNode(element.elem, xsd_type=xsd_type)
            self.assertEqual(typed_element.kind, 'element')

    def test_name_property(self):
        root = self.context.root
        attr = AttributeNode('a1', '20')
        namespace = NamespaceNode('xs', 'http://www.w3.org/2001/XMLSchema')

        self.assertEqual(root.name, 'root')
        self.assertEqual(attr.name, 'a1')
        self.assertEqual(namespace.name, 'xs')

        with patch.multiple(DummyXsdType, is_simple=lambda x: True):
            xsd_type = DummyXsdType()

            typed_elem = ElementNode(elem=root.elem, xsd_type=xsd_type)
            self.assertEqual(typed_elem.name, 'root')

            typed_attr = AttributeNode('a1', value='20', xsd_type=xsd_type)
            self.assertEqual(typed_attr.name, 'a1')

    def test_path_property(self):
        root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/><C2 max="10"/></B3></A>')

        context = XPathContext(root)

        self.assertEqual(context.root.path, '/A')
        self.assertEqual(context.root[0].path, '/A/B1')
        self.assertEqual(context.root[0][0].path, '/A/B1/C1')
        self.assertEqual(context.root[1].path, '/A/B2')
        self.assertEqual(context.root[2].path, '/A/B3')
        self.assertEqual(context.root[2][0].path, '/A/B3/C1')
        self.assertEqual(context.root[2][1].path, '/A/B3/C2')

        attr = context.root[2][1].attributes[0]
        self.assertEqual(attr.path, '/A/B3/C2/@max')

        document = ElementTree.ElementTree(root)
        context = XPathContext(root=document)
        self.assertEqual(context.root[0][2][0].path, '/A/B3/C1')

        root = ElementTree.XML('<A><B1>10</B1><B2 min="1"/><B3/></A>')
        context = XPathContext(root)
        with patch.object(DummyXsdType(), 'is_simple', return_value=True) as xsd_type:
            elem = context.root[0]
            elem.xsd_type = xsd_type
            self.assertEqual(elem.path, '/A/B1')

        with patch.object(DummyXsdType(), 'is_simple', return_value=True) as xsd_type:
            context = XPathContext(root)
            attr = context.root[1].attributes[0]
            attr.xsd_type = xsd_type
            self.assertEqual(attr.path, '/A/B2/@min')

    def test_element_node_iter(self):
        root = ElementTree.XML('<A>text1\n<B1 a="10">text2</B1><B2/><B3><C1>text3</C1></B3></A>')

        context = XPathContext(root)
        expected = [
            context.root, context.root.namespace_nodes[0],
            context.root[0],
            context.root[1], context.root[1].namespace_nodes[0],
            context.root[1].attributes[0], context.root[1][0],
            context.root[2], context.root[2].namespace_nodes[0],
            context.root[3], context.root[3].namespace_nodes[0],
            context.root[3][0], context.root[3][0].namespace_nodes[0],
            context.root[3][0][0]
        ]

        result = list(context.root.iter())
        self.assertListEqual(result, expected)

        root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/><C2/></B3></A>')
        context = XPathContext(root)

        # iter includes also xml namespace nodes
        self.assertListEqual(
            list(e.elem for e in context.root.iter() if isinstance(e, ElementNode)),
            list(root.iter())
        )

    def test_document_node_iter(self):
        root = ElementTree.XML('<A><B1><C1/></B1><B2/><B3><C1/><C2/></B3></A>')
        doc = ElementTree.ElementTree(root)
        context = XPathContext(doc)

        self.assertListEqual(
            list(e.elem for e in context.root.iter() if isinstance(e, ElementNode)),
            list(doc.iter())
        )

    @unittest.skipIf(lxml_etree is None, 'lxml.etree is not installed')
    def test_lazy_attributes_iter__issue_72(self):
        xml = lxml_etree.fromstring('<root id="test"></root>')
        node_tree = get_node_tree(root=xml)

        nodes = list(node for node in node_tree.iter_lazy())
        self.assertListEqual(nodes, [node_tree])

        nodes = list(node for node in node_tree.iter())
        self.assertListEqual(nodes, [
            node_tree, node_tree.namespace_nodes[0], node_tree.attributes[0]
        ])

        nodes = list(node for node in node_tree.iter_lazy())
        self.assertListEqual(nodes, [
            node_tree, node_tree.namespace_nodes[0], node_tree.attributes[0]
        ])

    def test_is_schema_node(self):
        root = ElementTree.XML('<root a="10">text</root>')
        context = XPathContext(root)

        self.assertFalse(context.root.is_schema_node())
        self.assertFalse(context.root.attributes[0].is_schema_node())
        self.assertIsNone(context.root.children[0].is_schema_node())

        if xmlschema is not None:
            schema = xmlschema.XMLSchema(dedent("""
                <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
                  <xs:element name="elem"/>
                  <xs:attribute name="attr" type="xs:string"/>
                </xs:schema>"""))

            context = XPathSchemaContext(schema)
            self.assertFalse(context.root.is_schema_node())  # Is the schema
            self.assertTrue(context.root.attributes[0].is_schema_node())
            self.assertTrue(context.root.children[0].is_schema_node())

    def test_etree_iter_paths(self):
        root = ElementTree.XML('<a><b1><c1/><c2/></b1><b2/><b3><c3/></b3></a>')
        root[2].append(ElementTree.Comment('a comment'))
        root[2].append(ElementTree.Element('c3'))  # duplicated tag

        items = list(etree_iter_paths(root))
        self.assertListEqual(items, [
            (root, '.'), (root[0], './Q{}b1[1]'),
            (root[0][0], './Q{}b1[1]/Q{}c1[1]'),
            (root[0][1], './Q{}b1[1]/Q{}c2[1]'),
            (root[1], './Q{}b2[1]'), (root[2], './Q{}b3[1]'),
            (root[2][0], './Q{}b3[1]/Q{}c3[1]'),
            (root[2][1], './Q{}b3[1]/comment()[1]'),
            (root[2][2], './Q{}b3[1]/Q{}c3[2]')
        ])
        self.assertListEqual(list(etree_iter_paths(root, path='')), [
            (root, ''), (root[0], 'Q{}b1[1]'),
            (root[0][0], 'Q{}b1[1]/Q{}c1[1]'),
            (root[0][1], 'Q{}b1[1]/Q{}c2[1]'),
            (root[1], 'Q{}b2[1]'), (root[2], 'Q{}b3[1]'),
            (root[2][0], 'Q{}b3[1]/Q{}c3[1]'),
            (root[2][1], 'Q{}b3[1]/comment()[1]'),
            (root[2][2], 'Q{}b3[1]/Q{}c3[2]')
        ])
        self.assertListEqual(list(etree_iter_paths(root, path='/')), [
            (root, '/'), (root[0], '/Q{}b1[1]'),
            (root[0][0], '/Q{}b1[1]/Q{}c1[1]'),
            (root[0][1], '/Q{}b1[1]/Q{}c2[1]'),
            (root[1], '/Q{}b2[1]'), (root[2], '/Q{}b3[1]'),
            (root[2][0], '/Q{}b3[1]/Q{}c3[1]'),
            (root[2][1], '/Q{}b3[1]/comment()[1]'),
            (root[2][2], '/Q{}b3[1]/Q{}c3[2]')
        ])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

def main() -> None:
    import xml.etree.ElementTree as ElementTree
    import lxml.etree as etree

    from typing import Iterator, Union, cast

    from xmlschema import XMLSchema
    from xmlschema.validators import XsdSimpleType, XsdComplexType, XsdAnyElement

    from elementpath.protocols import ElementProtocol, LxmlElementProtocol, \
        DocumentProtocol, LxmlDocumentProtocol, XsdTypeProtocol, XsdElementProtocol, \
        XsdAttributeProtocol, GlobalMapsProtocol, XsdSchemaProtocol

    ###
    # Test protocols for ElementTree and lxml.etree

    def iter_elements(element: ElementProtocol) -> Iterator[ElementProtocol]:
        for e in element.iter():
            yield e

    def iter_lxml_elements(element: LxmlElementProtocol) -> Iterator[LxmlElementProtocol]:
        for e in element.iter():
            yield e

    doc: DocumentProtocol
    elem: ElementProtocol
    lxml_doc: LxmlDocumentProtocol
    lxml_elem: LxmlElementProtocol

    doc = ElementTree.ElementTree()
    del doc

    elem = ElementTree.XML('<a><b1/><b2><c1/><c2/></b2><b3/></a>')

    elements = list(iter_elements(elem))
    elements.clear()

    lxml_doc = etree.ElementTree()
    del lxml_doc

    lxml_elem = etree.XML('<a><b1/><b2><c1/><c2/></b2><b3/></a>')

    lxml_elem2 = etree.XML('<a><b1/><b2><c1/><c2/></b2><b3/></a>')
    lxml_elem2 = lxml_elem
    del lxml_elem2

    elem2 = ElementTree.XML('<a><b1/><b2><c1/><c2/></b2><b3/></a>')
    elem2 = elem
    del elem2

    lxml_elements = list(iter_lxml_elements(lxml_elem))
    lxml_elements.clear()

    ###
    # Test protocols for XSD type annotations
    BaseXsdType = Union[XsdSimpleType, XsdComplexType]

    class Base:
        xsd_type: XsdTypeProtocol

        def __init__(self, xsd_type: XsdTypeProtocol) -> None:
            self.xsd_type = xsd_type

    class Derived(Base):
        def __init__(self, xsd_type: BaseXsdType) -> None:
            super().__init__(xsd_type)

    def check_elem_type(xsd_element: XsdElementProtocol) -> None:
        assert xsd_element.type is not None

    def check_any_elem_type(xsd_element: XsdElementProtocol) -> None:
        assert xsd_element.type is None

    def check_attr_type(xsd_attribute: XsdAttributeProtocol) -> bool:
        return xsd_attribute.type is not None

    def check_simple_type(xsd_type: XsdTypeProtocol) -> bool:
        return xsd_type.is_simple()

    def check_maps(maps: GlobalMapsProtocol) -> bool:
        return maps is not None

    def check_xsd_schema(s: XsdSchemaProtocol) -> None:
        assert s is not None

    schema = XMLSchema("""
        <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
            <xs:element name="elem1" type="xs:string"/>
            <xs:element name="elem2" type="xs:int"/>
            <xs:simpleType name="type1">
                <xs:restriction base="xs:string"/>
            </xs:simpleType>
            <xs:group name="group1">
                <xs:sequence>
                    <xs:any processContents="lax"/>
                </xs:sequence>
            </xs:group>
            <xs:attribute name="attr1" type="xs:int"/>
            <xs:attribute name="attr2" type="xs:float"/>
        </xs:schema>""")

    check_any_elem_type(cast(XsdAnyElement, schema.groups['group1'][0]))

    check_elem_type(schema.elements['elem1'])

    check_maps(schema.maps)

    check_xsd_schema(schema)

    a = cast(BaseXsdType, schema.types['type1'])
    check_simple_type(a)

    b = schema.attributes['attr1']
    check_attr_type(b)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# Copyright (c), 2022, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from timeit import timeit
from memory_profiler import profile

from elementpath import XPath1Parser
from elementpath.xpath_tokens import ValueToken


def run_timeit(stmt='pass', setup='pass', number=1000):
    seconds = timeit(stmt, setup=setup, number=number)
    print("{}: {}s".format(stmt, seconds))


@profile
def xpath_none_null_values():
    null_values = [None for _ in range(50000)]
    return null_values


@profile
def xpath_list_null_values():
    null_values = [[] for _ in range(50000)]
    return null_values


@profile
def xpath_tuple_null_values():
    null_values = [() for _ in range(50000)]
    return null_values


def is_empty_sequence(s):
    return not s and isinstance(s, list)


if __name__ == '__main__':
    print('*' * 68)
    print("*** Memory and timing profile of XPath null values alternatives  ***")
    print('*' * 68)
    print()

    NUMBER = 1000
    SETUP = 'from __main__ import obj1, obj2, is_empty_sequence'
    obj1 = []
    obj2 = ['foo', 'bar']

    print("*** Profile evaluation ***\n")

    run_timeit('[None for _ in range(10000)]', number=NUMBER)
    run_timeit('[[] for _ in range(10000)]', number=NUMBER)
    run_timeit('[() for _ in range(10000)]', number=NUMBER)

    print()

    run_timeit('for _ in range(10000): obj1 is None', SETUP, NUMBER)
    run_timeit('for _ in range(10000): obj1 == []', SETUP, NUMBER)
    run_timeit('for _ in range(10000): obj1 == ()', SETUP, NUMBER)
    run_timeit('for _ in range(10000): not obj1 and isinstance(obj1, list)', SETUP, NUMBER)
    run_timeit('for _ in range(10000): not obj1 and isinstance(obj1, tuple)', SETUP, NUMBER)
    run_timeit('for _ in range(10000): is_empty_sequence(obj1)', SETUP, NUMBER)

    print()

    run_timeit('for _ in range(10000): obj2 is None', SETUP, NUMBER)
    run_timeit('for _ in range(10000): obj2 == []', SETUP, NUMBER)
    run_timeit('for _ in range(10000): obj2 == ()', SETUP, NUMBER)
    run_timeit('for _ in range(10000): not obj2 and isinstance(obj2, list)', SETUP, NUMBER)
    run_timeit('for _ in range(10000): not obj2 and isinstance(obj2, tuple)', SETUP, NUMBER)
    run_timeit('for _ in range(10000): is_empty_sequence(obj2)', SETUP, NUMBER)

    print()
    print("*** Profile memory ***\n")

    xpath_none_null_values()
    xpath_list_null_values()
    xpath_tuple_null_values()



#
# Copyright (c), 2018-2020, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import re
import math
from calendar import isleap, leapdays
from decimal import Decimal
from operator import attrgetter
from typing import Any, List, Optional, Union, SupportsFloat
from urllib.parse import urlsplit

from elementpath._typing import Iterator, Match

###
# Common sets constants
OCCURRENCE_INDICATORS = frozenset(('?', '*', '+'))
BOOLEAN_VALUES = frozenset(('true', 'false', '1', '0'))
NUMERIC_INF_OR_NAN = frozenset(('INF', '-INF', 'NaN'))
INVALID_NUMERIC = frozenset(
    ('inf', '+inf', '-inf', 'nan', 'infinity', '+infinity', '-infinity')
)

###
# Data validation helpers

NORMALIZE_PATTERN = re.compile(r'[^\S\xa0]')
WHITESPACES_PATTERN = re.compile(r'[^\S\xa0]+')  # include ASCII 160 (non-breaking space)
NCNAME_PATTERN = re.compile(r'^[^\d\W][\w.\-\u00B7\u0300-\u036F\u203F\u2040]*$')
QNAME_PATTERN = re.compile(
    r'^(?:(?P<prefix>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*):)?'
    r'(?P<local>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*)$',
)
EQNAME_PATTERN = re.compile(
    r'^(?:Q{(?P<namespace>[^}]+)}|'
    r'(?P<prefix>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*):)?'
    r'(?P<local>[^\d\W][\w\-.\u00B7\u0300-\u036F\u0387\u06DD\u06DE\u203F\u2040]*)$',
)
WRONG_ESCAPE_PATTERN = re.compile(r'%(?![a-fA-F\d]{2})')
XML_NEWLINES_PATTERN = re.compile('\r\n|\r|\n')


def upper_camel_case(s: str) -> str:
    return re.sub(r'^\d+', '', re.sub(r'[\W_]', '', s.title()))


def collapse_white_spaces(s: str) -> str:
    return WHITESPACES_PATTERN.sub(' ', s).strip(' ')


def is_ncname(s: str) -> bool:
    return re.match(r'^[^\d\W][\w.\-\u00B7\u0300-\u036F\u203F\u2040]*$', s) is not None


def is_idrefs(value: Optional[str]) -> bool:
    return isinstance(value, str) and \
        all(NCNAME_PATTERN.match(x) is not None for x in value.split())


node_position = attrgetter('position')


###
# Date/Time helpers
MONTH_DAYS = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
MONTH_DAYS_LEAP = [0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]


def adjust_day(year: int, month: int, day: int) -> int:
    if month in (1, 3, 5, 7, 8, 10, 12):
        return day
    elif month in (4, 6, 9, 11):
        return min(day, 30)
    else:
        return min(day, 29) if isleap(year) else min(day, 28)


def days_from_common_era(year: int) -> int:
    """
    Returns the number of days from 0001-01-01 to the provided year. For a
    common era year the days are counted until the last day of December, for a
    BCE year the days are counted down from the end to the 1st of January.
    """
    if year > 0:
        return year * 365 + year // 4 - year // 100 + year // 400
    elif year >= -1:
        return year * 366
    else:
        year = -year - 1
        return -(366 + year * 365 + year // 4 - year // 100 + year // 400)


DAYS_IN_4Y = days_from_common_era(4)
DAYS_IN_100Y = days_from_common_era(100)
DAYS_IN_400Y = days_from_common_era(400)


def months2days(year: int, month: int, months_delta: int) -> int:
    """
    Converts a delta of months to a delta of days, counting from the 1st day of the month,
    relative to the year and the month passed as arguments.

    :param year: the reference start year, a negative or zero value means a BCE year \
    (0 is 1 BCE, -1 is 2 BCE, -2 is 3 BCE, etc.).
    :param month: the starting month (1-12).
    :param months_delta: the number of months, if negative count backwards.
    """
    if not months_delta:
        return 0

    total_months = month - 1 + months_delta
    target_year = year + total_months // 12
    target_month = total_months % 12 + 1

    if month <= 2:
        y_days = 365 * (target_year - year) + leapdays(year, target_year)
    else:
        y_days = 365 * (target_year - year) + leapdays(year + 1, target_year + 1)

    months_days = MONTH_DAYS_LEAP if isleap(target_year) else MONTH_DAYS
    if target_month >= month:
        m_days = sum(months_days[m] for m in range(month, target_month))
        return y_days + m_days if y_days >= 0 else y_days + m_days
    else:
        m_days = sum(months_days[m] for m in range(target_month, month))
        return y_days - m_days if y_days >= 0 else y_days - m_days


def round_number(value: Union[float, int, Decimal]) -> Union[float, int, Decimal]:
    if math.isnan(value) or math.isinf(value):
        return value

    number = Decimal(value)
    if number > 0:
        return type(value)(number.quantize(Decimal('1'), rounding='ROUND_HALF_UP'))
    else:
        return type(value)(number.quantize(Decimal('1'), rounding='ROUND_HALF_DOWN'))


def normalized_seconds(seconds: Union[int, Decimal]) -> str:
    # Decimal.normalize() does not remove exp every time: eg. Decimal('1E+1')
    return '{:.6f}'.format(seconds).rstrip('0').rstrip('.')


def is_xml_codepoint(cp: int) -> bool:
    return cp in (0x9, 0xA, 0xD) or \
        0x20 <= cp <= 0xD7FF or \
        0xE000 <= cp <= 0xFFFD or \
        0x10000 <= cp <= 0x10FFFF


def ordinal(n: int) -> str:
    if n in (11, 12, 13):
        return '%dth' % n

    least_significant_digit = n % 10
    if least_significant_digit == 1:
        return '%dst' % n
    elif least_significant_digit == 2:
        return '%dnd' % n
    elif least_significant_digit == 3:
        return '%drd' % n
    else:
        return '%dth' % n


def get_double(value: Union[SupportsFloat, str], xsd_version: str = '1.0') -> float:
    if isinstance(value, str):
        value = collapse_white_spaces(value)
        if value in NUMERIC_INF_OR_NAN or xsd_version != '1.0' and value == '+INF':
            if value == 'NaN':
                return math.nan  # for NaN use the predefined instance to keep identity
        elif value.lower() in INVALID_NUMERIC:
            raise ValueError(f'invalid value {value!r} for xs:double/xs:float')
    elif math.isnan(value):
        return math.nan

    return float(value)


def numeric_equal(op1: SupportsFloat, op2: SupportsFloat) -> bool:
    if op1 == op2:
        return True
    return math.isclose(op1, op2, rel_tol=1e-7, abs_tol=0.0)


def numeric_not_equal(op1: SupportsFloat, op2: SupportsFloat) -> bool:
    if op1 == op2:
        return False
    return not math.isclose(op1, op2, rel_tol=1e-7, abs_tol=0.0)


def equal(op1: Any, op2: Any) -> bool:
    if isinstance(op1, float) and math.isnan(op1):
        return isinstance(op2, float) and math.isnan(op2)
    return bool(op1 == op2)


def not_equal(op1: Any, op2: Any) -> bool:
    if isinstance(op1, float) and math.isnan(op1):
        return not isinstance(op2, float) or not math.isnan(op2)
    return bool(op1 != op2)


def match_wildcard(name: str, wildcard: str) -> bool:
    if wildcard == '*' or wildcard == '*:*':
        return True
    elif wildcard.startswith('*:'):
        if name.startswith('{'):
            return name.endswith(f'}}{wildcard[2:]}')
        else:
            return name == wildcard[2:]
    elif wildcard.startswith('{') and wildcard.endswith('}*') or wildcard.endswith(':*'):
        return name.startswith(wildcard[:-1])
    else:
        return False


def escape_json_string(s: str, escaped: bool = False) -> str:
    if escaped:
        s = s.replace('\\"', '"')
    else:
        s = s.replace('\\', '\\\\')

    s = s.replace('\"', '\\"').\
        replace('\b', r'\b').\
        replace('\r', r'\r').\
        replace('\n', r'\n').\
        replace('\t', r'\t').\
        replace('\f', r'\f').\
        replace('/', r'\/')
    return ''.join(
        rf'\u{ord(x):04X}' if 1 <= ord(x) <= 31 or 127 <= ord(x) <= 159 else x
        for x in s
    )


def unescape_json_string(s: str) -> str:

    def unicode_escape_callback(match: Match[str]) -> str:
        return chr(int(match.group(1).upper(), 16))

    s = s.replace('\\"', '\"').\
        replace(r'\b', '\b').\
        replace(r'\r', '\r').\
        replace(r'\n', '\n').\
        replace(r'\t', '\t').\
        replace(r'\f', '\f').\
        replace(r'\/', '/').\
        replace('\\\\', '\\')

    return re.sub(r'\\u([0-9A-Fa-f]{4})', unicode_escape_callback, s)


def iter_sequence(obj: Any) -> Iterator[Any]:
    if obj is None:
        return
    elif isinstance(obj, list):
        for item in obj:
            yield from iter_sequence(item)
    else:
        yield obj


def split_function_test(function_test: str) -> List[str]:
    if not function_test.startswith('function('):
        return []
    elif function_test == 'function(*)':
        return ['*']

    parts = function_test[9:].partition(') as ')
    if parts[0]:
        sequence_types = parts[0].split(', ')
        sequence_types.append(parts[2])
    else:
        sequence_types = [parts[2]]

    return sequence_types


def is_absolute_uri(uri: str) -> bool:
    try:
        parts = urlsplit(uri.strip())
    except ValueError:
        return False
    else:
        return parts.scheme == 'urn' or \
            parts.scheme != '' and parts.netloc != '' or \
            parts.path.startswith('/')



#!/usr/bin/env python
#
# Copyright (c), 2022, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
from textwrap import dedent
from xml.etree import ElementTree

try:
    import lxml.etree as lxml_etree
except ImportError:
    lxml_etree = None

try:
    import xmlschema
except ImportError:
    xmlschema = None

from elementpath.validators import validate_analyzed_string, validate_json_to_xml


@unittest.skipIf(xmlschema is None, "xmlschema library is not installed")
class ValidatorsTest(unittest.TestCase):
    etree = ElementTree

    def test_validate_analyzed_string(self):
        xml_source = dedent("""\
        <analyze-string-result xmlns="http://www.w3.org/2005/xpath-functions">
            <match>The</match>
            <non-match> </non-match>
            <match>cat</match>
            <non-match> </non-match>
            <match>sat</match>
            <non-match> </non-match>
            <match>on</match>
            <non-match> </non-match>
            <match>the</match>
            <non-match> </non-match>
            <match>mat</match>
            <non-match>.</non-match>
        </analyze-string-result>
        """)

        root = self.etree.XML(xml_source)
        self.assertIsNone(validate_analyzed_string(root))

        with self.assertRaises(ValueError):
            validate_analyzed_string(self.etree.XML('<invalid/>'))

    def test_validate_json_to_xml(self):
        xml_source = """\
        <fn:array xmlns:fn="http://www.w3.org/2005/xpath-functions">
           <fn:number>1</fn:number>
        </fn:array>"""

        root = self.etree.XML(xml_source)
        self.assertIsNone(validate_json_to_xml(root))

        with self.assertRaises(ValueError):
            validate_json_to_xml(self.etree.XML('<invalid/>'))


@unittest.skipIf(xmlschema is None, "xmlschema library is not installed")
@unittest.skipIf(lxml_etree is None, "lxml library is not installed")
class ValidatorsTest(ValidatorsTest):
    etree = lxml_etree


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c), 2024, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XSD atomic datatypes subpackage. Includes a class for UntypedAtomic data and
classes for other XSD built-in types. This subpackage raises only built-in
exceptions in order to be reusable in other packages.
"""
from collections import namedtuple
from decimal import Decimal
from typing import Any, Optional, Type, TYPE_CHECKING

from elementpath._typing import MutableMapping
from elementpath.aliases import AnyNsmapType
from elementpath.protocols import XsdTypeProtocol
from elementpath.exceptions import xpath_error
from elementpath.namespaces import XSD_NAMESPACE, XSD_NOTATION, XSD_UNTYPED_ATOMIC

import elementpath.datatypes as dt

if TYPE_CHECKING:
    from elementpath.xpath_tokens import XPathToken

Builder = namedtuple('Builder', ['cls', 'value'])

ATOMIC_BUILDERS: MutableMapping[Optional[str], Builder] = {
    f'{{{XSD_NAMESPACE}}}untypedAtomic': Builder(dt.UntypedAtomic, '1'),
    f'{{{XSD_NAMESPACE}}}anyType': Builder(dt.UntypedAtomic, '1'),
    f'{{{XSD_NAMESPACE}}}anySimpleType': Builder(dt.UntypedAtomic, '1'),
    f'{{{XSD_NAMESPACE}}}anyAtomicType': Builder(dt.UntypedAtomic, '1'),
    f'{{{XSD_NAMESPACE}}}boolean': Builder(bool, True),
    f'{{{XSD_NAMESPACE}}}decimal': Builder(Decimal, '1.0'),
    f'{{{XSD_NAMESPACE}}}double': Builder(float, 1.0),
    f'{{{XSD_NAMESPACE}}}float': Builder(dt.Float10, 1.0),
    f'{{{XSD_NAMESPACE}}}string': Builder(str, '  alpha\t'),
    f'{{{XSD_NAMESPACE}}}date': Builder(dt.Date, '2000-01-01'),
    f'{{{XSD_NAMESPACE}}}dateTime': Builder(dt.DateTime, '2000-01-01T12:00:00'),
    f'{{{XSD_NAMESPACE}}}gDay': Builder(dt.GregorianDay, '---31'),
    f'{{{XSD_NAMESPACE}}}gMonth': Builder(dt.GregorianMonth, '--12'),
    f'{{{XSD_NAMESPACE}}}gMonthDay': Builder(dt.GregorianMonthDay, '--12-01'),
    f'{{{XSD_NAMESPACE}}}gYear': Builder(dt.GregorianYear, '1999'),
    f'{{{XSD_NAMESPACE}}}gYearMonth': Builder(dt.GregorianYearMonth, '1999-09'),
    f'{{{XSD_NAMESPACE}}}time': Builder(dt.Time, '09:26:54'),
    f'{{{XSD_NAMESPACE}}}duration': Builder(dt.Duration, 'P1MT1S'),
    f'{{{XSD_NAMESPACE}}}dayTimeDuration': Builder(dt.DayTimeDuration, 'P1DT1S'),
    f'{{{XSD_NAMESPACE}}}yearMonthDuration': Builder(dt.YearMonthDuration, 'P1Y1M'),
    f'{{{XSD_NAMESPACE}}}QName':
        Builder(dt.QName, ("http://www.w3.org/2001/XMLSchema", 'xs:element')),
    f'{{{XSD_NAMESPACE}}}anyURI': Builder(dt.AnyURI, 'https://example.com'),
    f'{{{XSD_NAMESPACE}}}normalizedString': Builder(dt.NormalizedString, ' alpha  '),
    f'{{{XSD_NAMESPACE}}}token': Builder(dt.XsdToken, 'a token'),
    f'{{{XSD_NAMESPACE}}}language': Builder(dt.Language, 'en-US'),
    f'{{{XSD_NAMESPACE}}}Name': Builder(dt.Name, '_a.name::'),
    f'{{{XSD_NAMESPACE}}}NCName': Builder(dt.NCName, 'nc-name'),
    f'{{{XSD_NAMESPACE}}}ID': Builder(dt.Id, 'id1'),
    f'{{{XSD_NAMESPACE}}}IDREF': Builder(dt.Idref, 'id_ref1'),
    f'{{{XSD_NAMESPACE}}}ENTITY': Builder(dt.Entity, 'entity1'),
    f'{{{XSD_NAMESPACE}}}NMTOKEN': Builder(dt.NMToken, 'a_token'),
    f'{{{XSD_NAMESPACE}}}base64Binary': Builder(dt.Base64Binary, b'YWxwaGE='),
    f'{{{XSD_NAMESPACE}}}hexBinary': Builder(dt.HexBinary, b'31'),
    f'{{{XSD_NAMESPACE}}}dateTimeStamp':
        Builder(dt.DateTimeStamp.fromstring, '2000-01-01T12:00:00+01:00'),
    f'{{{XSD_NAMESPACE}}}integer': Builder(dt.Integer, 1),
    f'{{{XSD_NAMESPACE}}}long': Builder(dt.Long, 1),
    f'{{{XSD_NAMESPACE}}}int': Builder(dt.Int, 1),
    f'{{{XSD_NAMESPACE}}}short': Builder(dt.Short, 1),
    f'{{{XSD_NAMESPACE}}}byte': Builder(dt.Byte, 1),
    f'{{{XSD_NAMESPACE}}}positiveInteger': Builder(dt.PositiveInteger, 1),
    f'{{{XSD_NAMESPACE}}}negativeInteger': Builder(dt.NegativeInteger, -1),
    f'{{{XSD_NAMESPACE}}}nonPositiveInteger': Builder(dt.NonPositiveInteger, 0),
    f'{{{XSD_NAMESPACE}}}nonNegativeInteger': Builder(dt.NonNegativeInteger, 0),
    f'{{{XSD_NAMESPACE}}}unsignedLong': Builder(dt.UnsignedLong, 1),
    f'{{{XSD_NAMESPACE}}}unsignedInt': Builder(dt.UnsignedInt, 1),
    f'{{{XSD_NAMESPACE}}}unsignedShort': Builder(dt.UnsignedShort, 1),
    f'{{{XSD_NAMESPACE}}}unsignedByte': Builder(dt.UnsignedByte, 1),
}


def get_builder(xsd_type: Optional[XsdTypeProtocol],
                text: Optional[str],
                namespaces: AnyNsmapType = None) -> Builder:
    """
    Returns the atomic builtin XSD type that is suitable for decoding a text value.
    """
    if xsd_type is None:
        return ATOMIC_BUILDERS[XSD_UNTYPED_ATOMIC]
    elif xsd_type.name in ATOMIC_BUILDERS:
        return ATOMIC_BUILDERS[xsd_type.name]

    if namespaces is None:
        namespaces = {}

    root_type = xsd_type.root_type
    derivation_depth = 1

    while root_type.name not in ATOMIC_BUILDERS or root_type.is_union():
        derivation_depth += 1
        if derivation_depth > 15:  # too many derivations: it could be a dummy XSD type
            return ATOMIC_BUILDERS[XSD_UNTYPED_ATOMIC]

        union_type = root_type
        if hasattr(union_type, 'member_types'):
            for xsd_type in union_type.member_types:
                assert xsd_type is not None
                if xsd_type.root_type.is_valid(text, namespaces=namespaces):
                    root_type = xsd_type.root_type
                    break
            else:
                if text is None:
                    return ATOMIC_BUILDERS[XSD_UNTYPED_ATOMIC]

    if root_type.is_notation():
        if root_type.name is None or root_type.name == XSD_NOTATION:
            name = '_Notation'
        elif '}' in root_type.name:
            name = f"{root_type.name.split('}')[-1].title()}_Notation"
        else:
            name = f"{root_type.name.split(':')[-1].title()}_Notation"

        return Builder(type(name, (dt.Notation,), {}), ATOMIC_BUILDERS[root_type.name])

    return ATOMIC_BUILDERS[root_type.name]


def get_atomic_value(xsd_type: Optional[XsdTypeProtocol] = None,
                     text: Optional[str] = None,
                     namespaces: AnyNsmapType = None,
                     token: Optional['XPathToken'] = None) -> dt.AtomicType:
    """Gets an atomic value for an XSD type instance and a source text/value."""

    def decode_value(v: Any) -> dt.AtomicType:
        if isinstance(v, str) and issubclass(cls, (dt.QName, dt.Notation)) and namespaces:
            if ':' not in v:
                v = namespaces.get(''), v
            else:
                v = namespaces[v.split(':')[0]], v

        try:
            if isinstance(v, tuple):
                return cls(*v)
            elif issubclass(cls, (dt.AbstractDateTime, dt.Duration)):
                return cls.fromstring(v)
            return cls(v)
        except ValueError as err:
            raise xpath_error('FORG0001', err, token, namespaces)
        except ArithmeticError as err:
            if issubclass(cls, dt.AbstractDateTime):
                raise xpath_error('FODT0001', err, token, namespaces)
            elif issubclass(cls, dt.Duration):
                raise xpath_error('FODT0002', err, token, namespaces)
            else:
                raise xpath_error('FOCA0002', err, token, namespaces)

    if xsd_type is None:
        return dt.UntypedAtomic(text or '')

    if namespaces is None:
        namespaces = {}

    builder = get_builder(xsd_type, text, namespaces)
    cls: Type[dt.AtomicType] = builder.cls

    if text is None:
        return decode_value(builder.value)

    xsd_type.validate(text, namespaces=namespaces)
    return decode_value(text)


__all__ = ['get_atomic_value']



#!/usr/bin/env python
#
# Copyright (c), 2022, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import unittest
import io
from textwrap import dedent
from xml.etree import ElementTree

try:
    import xmlschema
except ImportError:
    xmlschema = None

from elementpath.sequence_types import normalize_sequence_type, is_instance, \
    is_sequence_type, match_sequence_type, is_sequence_type_restriction
from elementpath import XPath2Parser, XPathContext
from elementpath.xpath3 import XPath30Parser, XPath31Parser
from elementpath.namespaces import XSD_NAMESPACE, XSD_UNTYPED_ATOMIC, \
    XSD_ANY_ATOMIC_TYPE, XSD_ANY_SIMPLE_TYPE, XSI_NIL, XSD_STRING
from elementpath.datatypes import UntypedAtomic
from elementpath.xpath_nodes import CommentNode


class SequenceTypesTest(unittest.TestCase):

    def test_normalize_sequence_type_function(self):
        self.assertEqual(normalize_sequence_type(' xs:integer + '), 'xs:integer+')
        self.assertEqual(normalize_sequence_type(' xs :integer + '), 'xs :integer+')  # Invalid
        self.assertEqual(normalize_sequence_type(' element( * ) '), 'element(*)')
        self.assertEqual(normalize_sequence_type(' element( *,xs:int ) '), 'element(*, xs:int)')
        self.assertEqual(normalize_sequence_type(' \nfunction  ( * )\t '), 'function(*)')
        self.assertEqual(
            normalize_sequence_type(' \nfunction  ( item( ) * ) as  xs:integer\t '),
            'function(item()*) as xs:integer'
        )

    def test_is_sequence_type_restriction_function(self):
        self.assertTrue(is_sequence_type_restriction('xs:error?', 'none'))
        self.assertTrue(is_sequence_type_restriction('empty-sequence()', 'none'))
        self.assertTrue(is_sequence_type_restriction('xs:error*', 'empty-sequence()'))

        self.assertFalse(is_sequence_type_restriction('xs:integer', 'item()*'))
        self.assertFalse(is_sequence_type_restriction('xs:integer', 'item()?'))
        self.assertFalse(is_sequence_type_restriction('xs:integer', 'item()'))

        self.assertFalse(is_sequence_type_restriction('xs:integer+', 'xs:integer*'))
        self.assertFalse(is_sequence_type_restriction('xs:integer+', 'xs:integer?'))
        self.assertTrue(is_sequence_type_restriction('xs:integer+', 'xs:integer+'))
        self.assertTrue(is_sequence_type_restriction('xs:integer+', 'xs:integer'))

        self.assertFalse(is_sequence_type_restriction('xs:integer*', 'xs:integer?'))
        self.assertFalse(is_sequence_type_restriction('xs:integer*', 'xs:integer+'))
        self.assertTrue(is_sequence_type_restriction('xs:integer*', 'xs:integer*'))
        self.assertTrue(is_sequence_type_restriction('xs:integer*', 'xs:integer'))

        self.assertFalse(is_sequence_type_restriction('xs:integer?', 'xs:integer*'))
        self.assertFalse(is_sequence_type_restriction('xs:integer?', 'xs:integer+'))
        self.assertTrue(is_sequence_type_restriction('xs:integer?', 'xs:integer?'))
        self.assertTrue(is_sequence_type_restriction('xs:integer?', 'xs:integer'))

        self.assertTrue(is_sequence_type_restriction('node()', 'element()'))
        self.assertFalse(is_sequence_type_restriction('element()', 'node()'))
        self.assertTrue(is_sequence_type_restriction('xs:anyAtomicType', 'xs:string'))
        self.assertFalse(is_sequence_type_restriction('xs:anyAtomicType', 'xs:unknown'))
        self.assertTrue(is_sequence_type_restriction('xs:string', 'xs:anyAtomicType'))
        self.assertTrue(is_sequence_type_restriction('xs:string', 'xs:token'))
        self.assertFalse(is_sequence_type_restriction('xs:string', 'xs:int'))
        self.assertFalse(is_sequence_type_restriction('xs:string', 'xs:unknown'))
        self.assertFalse(is_sequence_type_restriction('element()', 'xs:string'))
        self.assertFalse(is_sequence_type_restriction('function(*)', 'xs:string'))
        self.assertFalse(is_sequence_type_restriction(
            'function(item()+) as xs:boolean', 'function(item()*) as xs:boolean'
        ))
        self.assertTrue(is_sequence_type_restriction(
            'function(item()) as xs:boolean', 'function(item()*) as xs:boolean'
        ))
        self.assertFalse(is_sequence_type_restriction(
            'function(item()+) as xs:boolean', 'function(item()) as xs:boolean'
        ))
        self.assertFalse(is_sequence_type_restriction(
            'function(item()) as xs:boolean', 'function(item()) as xs:string'
        ))
        self.assertFalse(is_sequence_type_restriction(
            'function(item(), item()) as xs:boolean', 'function(item()) as xs:boolean'
        ))

    def test_is_instance_function(self):
        self.assertTrue(is_instance(UntypedAtomic(1), XSD_UNTYPED_ATOMIC))
        self.assertFalse(is_instance(1, XSD_UNTYPED_ATOMIC))
        self.assertTrue(is_instance(1, XSD_ANY_ATOMIC_TYPE))
        self.assertFalse(is_instance([1], XSD_ANY_ATOMIC_TYPE))
        self.assertTrue(is_instance(1, XSD_ANY_SIMPLE_TYPE))
        self.assertTrue(is_instance([1], XSD_ANY_SIMPLE_TYPE))

        self.assertTrue(is_instance('foo', '{%s}string' % XSD_NAMESPACE))
        self.assertFalse(is_instance(1, '{%s}string' % XSD_NAMESPACE))
        self.assertTrue(is_instance(1.0, '{%s}double' % XSD_NAMESPACE))
        self.assertFalse(is_instance(1.0, '{%s}float' % XSD_NAMESPACE))

        parser = XPath2Parser(xsd_version='1.1')
        self.assertTrue(is_instance(1.0, '{%s}double' % XSD_NAMESPACE), parser)
        self.assertFalse(is_instance(1.0, '{%s}float' % XSD_NAMESPACE), parser)

        self.assertRaises(KeyError, is_instance, 'foo', '{%s}unknown' % XSD_NAMESPACE)
        self.assertRaises(KeyError, is_instance, 'foo', '{%s}unknown' % XSD_NAMESPACE, parser)

        self.assertRaises(KeyError, is_instance, 'foo', 'tst:unknown')
        self.assertRaises(KeyError, is_instance, 'foo', 'tst:unknown', parser)

        self.assertTrue(is_instance(None, '{%s}error' % XSD_NAMESPACE))
        self.assertTrue(is_instance([], '{%s}error' % XSD_NAMESPACE))
        self.assertFalse(is_instance(1.0, '{%s}error' % XSD_NAMESPACE))

        self.assertTrue(is_instance(1.0, '{%s}numeric' % XSD_NAMESPACE))
        self.assertFalse(is_instance(True, '{%s}numeric' % XSD_NAMESPACE))
        self.assertFalse(is_instance('foo', '{%s}numeric' % XSD_NAMESPACE))

    def test_is_sequence_type_function(self):
        self.assertTrue(is_sequence_type('empty-sequence()'))
        self.assertTrue(is_sequence_type('xs:string'))
        self.assertTrue(is_sequence_type('xs:float+'))
        self.assertTrue(is_sequence_type('element()*'))
        self.assertTrue(is_sequence_type('item()?'))
        self.assertTrue(is_sequence_type('xs:untypedAtomic+'))

        self.assertFalse(is_sequence_type(10))
        self.assertFalse(is_sequence_type(''))
        self.assertFalse(is_sequence_type('empty-sequence()*'))
        self.assertFalse(is_sequence_type('unknown'))
        self.assertFalse(is_sequence_type('unknown?'))
        self.assertFalse(is_sequence_type('tns0:unknown'))

        self.assertTrue(is_sequence_type(' element( ) '))
        self.assertTrue(is_sequence_type(' element( * ) '))
        self.assertFalse(is_sequence_type(' element( *, * ) '))
        self.assertTrue(is_sequence_type('element(A)'))
        self.assertTrue(is_sequence_type('element(A, xs:date)'))
        self.assertTrue(is_sequence_type('element(*, xs:date)'))
        self.assertFalse(is_sequence_type('element(A, B, xs:date)'))

        self.assertTrue(is_sequence_type('document-node(element(*, xs:date))'))
        self.assertFalse(is_sequence_type('document-node(element(*, xs:date)'))
        self.assertFalse(is_sequence_type('document-node(xs:date)'))

        parser = XPath2Parser()
        self.assertFalse(is_sequence_type('function(*)', parser))
        self.assertFalse(is_sequence_type('function(xs:string)', parser))
        self.assertFalse(is_sequence_type('map(xs:string, xs:string)', parser))
        self.assertFalse(is_sequence_type('array(xs:string)', parser))

        parser = XPath30Parser()
        self.assertTrue(is_sequence_type('function(*)', parser))
        self.assertTrue(is_sequence_type('function(xs:string)', parser))
        self.assertFalse(is_sequence_type('function(xs:string', parser))
        self.assertFalse(is_sequence_type('map(xs:string, xs:string)', parser))
        self.assertFalse(is_sequence_type('array(xs:string)', parser))

        parser = XPath31Parser()
        self.assertTrue(is_sequence_type('function(*)', parser))
        self.assertTrue(is_sequence_type('map(xs:string, xs:string)', parser))
        self.assertFalse(is_sequence_type('map(xs:string, xs:string', parser))
        self.assertTrue(is_sequence_type('array(xs:string)', parser))

        # Without a parser argument assumes the latest version coverage
        self.assertTrue(is_sequence_type('function(*)'))
        self.assertTrue(is_sequence_type('map(xs:string, xs:string)'))
        self.assertFalse(is_sequence_type('map(xs:string, xs:string'))
        self.assertTrue(is_sequence_type('array(xs:string)'))

        self.assertTrue(is_sequence_type('function(xs:int) as xs:int'))
        self.assertFalse(is_sequence_type('function(xs:unknown) as xs:int'))
        self.assertFalse(is_sequence_type('function(xs:int) as xs:unknown'))
        self.assertTrue(is_sequence_type('function(xs:int, ...) as xs:int'))

        self.assertTrue(is_sequence_type('function(xs:int, function(*)) as xs:int'))
        self.assertTrue(is_sequence_type('function(function(*), xs:int) as xs:int'))

        self.assertTrue(
            is_sequence_type('function(xs:int, function(xs:int) as xs:int) as xs:int')
        )
        self.assertFalse(
            is_sequence_type('function(function(xs:int) as xs:int, xs:int) as xs:int')
        )

    def test_match_sequence_type_function(self):
        self.assertTrue(match_sequence_type(None, 'empty-sequence()'))
        self.assertTrue(match_sequence_type([], 'empty-sequence()'))
        self.assertFalse(match_sequence_type('', 'empty-sequence()'))

        self.assertFalse(match_sequence_type('', 'empty-sequence()'))

        context = XPathContext(ElementTree.XML('<root><e1>1</e1><e2/><e3/></root>'))
        root = context.root

        self.assertTrue(match_sequence_type(root, 'element()'))
        self.assertTrue(match_sequence_type(root, 'element(root)'))
        self.assertFalse(match_sequence_type(root, 'element(foo)'))

        self.assertTrue(match_sequence_type(root, 'element(root, xs:untyped)'))
        self.assertTrue(match_sequence_type(root, 'element(root, xs:untyped?)'))

        if xmlschema is not None:
            schema = xmlschema.XMLSchema(dedent('''\
                <xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
                  <xs:element name="root" type="xs:string"/>
                </xs:schema>'''))

            root.xsd_type = schema.maps.types[XSD_STRING]
            self.assertFalse(match_sequence_type(root, 'element(root, xs:untyped)'))
            root.xsd_type = None

        root.elem.attrib[XSI_NIL] = 'true'
        self.assertFalse(match_sequence_type(root, 'element(root, xs:untyped)'))
        self.assertTrue(match_sequence_type(root, 'element(root, xs:untyped?)'))
        root.elem.attrib.pop(XSI_NIL)

        self.assertFalse(match_sequence_type(root, 'element(root, xs:string)'))
        self.assertFalse(match_sequence_type(root, 'element(root/e1, xs:string)'))
        self.assertTrue(match_sequence_type(root, 'element(root, xs:untypedAtomic)'))
        self.assertFalse(match_sequence_type(root, 'element(xs:root, xs:untypedAtomic)'))

        with self.assertRaises(NameError):
            match_sequence_type(root, 'element(root, xs:unknown)')

        parser = XPath2Parser()
        self.assertFalse(match_sequence_type(root, 'element(xs:root)', parser))
        self.assertFalse(match_sequence_type(root, 'element(tns:root)', parser))

        self.assertFalse(match_sequence_type(1.0, 'element()'))
        self.assertTrue(match_sequence_type([root], 'element()'))
        self.assertTrue(match_sequence_type(root, 'element()?'))
        self.assertTrue(match_sequence_type(root, 'element()+'))
        self.assertTrue(match_sequence_type(root, 'element()*'))
        self.assertFalse(match_sequence_type(root[:], 'element()'))
        self.assertFalse(match_sequence_type(root[:], 'element()?'))
        self.assertTrue(match_sequence_type(root[:], 'element()+'))
        self.assertTrue(match_sequence_type(root[:], 'element()*'))

        self.assertTrue(match_sequence_type(root, 'element(*)'))

        document = ElementTree.parse(io.StringIO('<root/>'))
        context = XPathContext(document)
        root = context.root

        self.assertTrue(match_sequence_type(root, 'document-node(element())'))
        self.assertFalse(match_sequence_type(root, 'document-node(element(A))'))

        parser = XPath2Parser()
        self.assertTrue(match_sequence_type(UntypedAtomic(1), 'xs:untypedAtomic'))
        self.assertFalse(match_sequence_type(1, 'xs:untypedAtomic'))

        self.assertTrue(match_sequence_type('1', 'xs:string'))
        self.assertFalse(match_sequence_type(1, 'xs:string'))

        with self.assertRaises(NameError) as ctx:
            match_sequence_type('1', 'xs:unknown', parser)
        self.assertIn('XPST0051', str(ctx.exception))

        with self.assertRaises(NameError) as ctx:
            match_sequence_type('1', 'tns0:string', parser)
        self.assertIn('XPST0051', str(ctx.exception))

        token = parser.parse('true()')
        self.assertFalse(match_sequence_type(1.0, 'function(*)'))
        self.assertTrue(match_sequence_type(token, 'function(*)'))
        self.assertTrue(match_sequence_type(token, 'function() as xs:boolean'))
        self.assertFalse(match_sequence_type(token, 'function() as xs:int'))

        parser = XPath31Parser()
        self.assertFalse(match_sequence_type(1.0, 'array(*)'))
        self.assertFalse(match_sequence_type(1.0, 'map(*)'))

        token = parser.parse('[1, 2, 3]')
        self.assertTrue(match_sequence_type(token, 'array(*)'))
        self.assertTrue(match_sequence_type(token, 'array(xs:integer)'))
        self.assertFalse(match_sequence_type(token, 'array(xs:string)'))
        self.assertFalse(match_sequence_type(token, 'map(*)'))

        token = parser.parse('map{1: 2}')
        self.assertFalse(match_sequence_type(token, 'array(*)'))
        self.assertTrue(match_sequence_type(token, 'map(*)'))
        self.assertTrue(match_sequence_type(token, 'map(xs:integer, xs:integer)'))
        self.assertFalse(match_sequence_type(token, 'map(xs:string, xs:integer)'))

        with self.assertRaises(SyntaxError):
            match_sequence_type(token, 'map(xs:integer+, xs:integer)')

        self.assertFalse(match_sequence_type('foo', 'xs:anyURI'))
        self.assertTrue(match_sequence_type('foo', 'xs:anyURI', strict=False))

        comment = ElementTree.Comment('foo')
        comment_node = CommentNode(comment)
        self.assertTrue(match_sequence_type(comment_node, 'comment()'))
        self.assertFalse(match_sequence_type(comment_node, 'comment(*)'))


if __name__ == '__main__':
    unittest.main()
